/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "App.h"
#include "Defs.h"
#include "Frame.h"

#include "SiteList.h"
#include "TownList.h"

#include "GlobeCache.h"
#include "GTOPO30Cache.h"
#include "SRTMCache.h"

#include <wx/config.h>

const wxString VENDOR_NAME = wxT("G4KLX");
const wxString APP_NAME    = wxT("ukwtools");

const wxString SECTION_GENERAL  = wxT("General");
const wxString SECTION_3D_MAP   = wxT("3D Map");
const wxString SECTION_COVERAGE = wxT("Coverage");
const wxString SECTION_HORIZON  = wxT("Horizon");
const wxString SECTION_MAP      = wxT("Map");

const wxString KEY_SRTM_PATH    = wxT("srtm_path");
const wxString KEY_GLOBE_PATH   = wxT("globe_path");
const wxString KEY_GTOPO30_PATH = wxT("gtopo30_path");
const wxString KEY_DTEM         = wxT("dtem");
const wxString KEY_DATA_PATH    = wxT("data_path");
const wxString KEY_MAP_QTH_PATH = wxT("map_qth_path");
const wxString KEY_MAP_SIZE     = wxT("map_size");
const wxString KEY_HEIGHT       = wxT("height");
const wxString KEY_FREQUENCY    = wxT("frequency");
const wxString KEY_TX_POWER     = wxT("tx_power");
const wxString KEY_TX_ANT_GAIN  = wxT("tx_ant_gain");
const wxString KEY_RX_MIN_SIG   = wxT("rx_min_signal");
const wxString KEY_RX_ANT_GAIN  = wxT("rx_ant_gain");
const wxString KEY_POLARISATION = wxT("polarisation");
const wxString KEY_CLIMATE      = wxT("climate");
const wxString KEY_GROUND       = wxT("ground");
const wxString KEY_CONFIDENCE   = wxT("confidence");
const wxString KEY_DISTANCE     = wxT("distance");
const wxString KEY_MIN_ELE      = wxT("min_eleation");
const wxString KEY_MAX_ELE      = wxT("max_elevation");
const wxString KEY_K_VALUE      = wxT("k");
const wxString KEY_FORMAT       = wxT("format");
const wxString KEY_RESOLUTION   = wxT("resolution");
const wxString KEY_TYPE         = wxT("type");
const wxString KEY_TOWNS        = wxT("towns");
const wxString KEY_SITES        = wxT("sites");
const wxString KEY_TEXT_SIZE    = wxT("text_size");

#if defined(__UNIX__)
const wxString DEFAULT_SRTM_PATH    = wxT("/usr/share/ukwtools/srtm");
const wxString DEFAULT_GLOBE_PATH   = wxT("/usr/share/ukwtools/globe");
const wxString DEFAULT_GTOPO30_PATH = wxT("/usr/share/ukwtools/gtopo30");
const wxString DEFAULT_DATA_PATH    = wxT("/usr/share/ukwtools");
const wxString DEFAULT_MAP_QTH_PATH = DEFAULT_DATA_PATH;
#else
const wxString DEFAULT_SRTM_PATH    = wxT("C:\\Program Files\\ukwtools\\data\\srtm");
const wxString DEFAULT_GLOBE_PATH   = wxT("C:\\Program Files\\ukwtools\\data\\globe");
const wxString DEFAULT_GTOPO30_PATH = wxT("C:\\Program Files\\ukwtools\\data\\gtopo30");
const wxString DEFAULT_DATA_PATH    = wxT("C:\\Program Files\\ukwtools\\data");
const wxString DEFAULT_MAP_QTH_PATH = DEFAULT_DATA_PATH;
#endif

const long    DEFAULT_DTEM         = long(DTEM_SRTM_GTOPO30);

const long    DEFAULT_MAP_SIZE     = 30L;
const long    DEFAULT_DHEIGHT      = 10L;
const long    DEFAULT_DISTANCE     = 50L;
const long    DEFAULT_FREQUENCY    = 2400L;
const long    DEFAULT_TX_POWER     = 5L;
const long    DEFAULT_TX_ANT_GAIN  = 15L;
const long    DEFAULT_RX_MIN_SIG   = -83L;
const long    DEFAULT_RX_ANT_GAIN  = 15L;
const long    DEFAULT_POLARISATION = 1L;
const long    DEFAULT_CLIMATE      = 5L;
const long    DEFAULT_GROUND       = 5L;
const long    DEFAULT_CONFIDENCE   = 90L;
const double  DEFAULT_MIN_ELE      = -2.0;
const double  DEFAULT_MAX_ELE      = 2.5;
const double  DEFAULT_K_VALUE      = 1.33;
const EFormat DEFAULT_FORMAT       = FORMAT_LOCATOR;
const long    DEFAULT_RESOLUTION   = 250L;
const long    DEFAULT_TYPE         = 0L;
const bool    DEFAULT_TOWNS        = true;
const bool    DEFAULT_SITES        = true;
const long    DEFAULT_TEXT_SIZE    = 6L;

IMPLEMENT_APP(CApp)

CApp::CApp() :
wxApp(),
m_cache(NULL),
m_fileName(),
m_centre(),
m_resolution(0),
m_towns(false),
m_sites(false),
m_points()
{
	for (int i = 0; i < MAX_POINT_COUNT; i++)
		m_points[i] = NULL;
}

CApp::~CApp()
{
	for (int i = 0; i < MAX_POINT_COUNT; i++) {
		if (m_points[i] != NULL)
			delete m_points[i];
	}
}

bool CApp::OnInit()
{
	SetVendorName(VENDOR_NAME);
	SetAppName(APP_NAME);

	wxConfigBase* config = wxConfigBase::Get();
	config->SetRecordDefaults();

	wxString srtmPath, globePath, gtopo30Path, dataPath, mapQTHPath;
	EDTEM dtem;
	getGeneralPreferences(srtmPath, globePath, gtopo30Path, dtem, dataPath, mapQTHPath);

	switch (dtem) {
		case DTEM_GLOBE_V1:
			m_cache = new CGlobeCache(globePath, DTEM_GLOBE_V1);
			break;

		case DTEM_GLOBE_V2:
			m_cache = new CGlobeCache(globePath, DTEM_GLOBE_V2);
			break;

		case DTEM_GTOPO30:
			m_cache = new CGTOPO30Cache(gtopo30Path);
			break;

		case DTEM_SRTM:
			m_cache = new CSRTMCache(srtmPath);
			break;

		case DTEM_SRTM_GLOBE_V1: {
			CGlobeCache* cache = new CGlobeCache(globePath, DTEM_GLOBE_V1);
			m_cache = new CSRTMCache(srtmPath, cache);
			break;
		}

		case DTEM_SRTM_GLOBE_V2: {
			CGlobeCache* cache = new CGlobeCache(globePath, DTEM_GLOBE_V2);
			m_cache = new CSRTMCache(srtmPath, cache);
			break;
		}

		case DTEM_SRTM_GTOPO30: {
			CGTOPO30Cache* cache = new CGTOPO30Cache(gtopo30Path);
			m_cache = new CSRTMCache(srtmPath, cache);
			break;
		}
	}

	CSiteList::create(dataPath);
	CTownList::create(dataPath);

	wxFrame* frame = new CFrame;
	frame->Show();

	SetTopWindow(frame);

	return true;
}

int CApp::OnExit()
{
	wxASSERT(m_cache != NULL);

	delete wxConfigBase::Set(NULL);

	delete m_cache;

	return 0;
}

void CApp::getGeneralPreferences(wxString& srtmPath, wxString& globePath, wxString& gtopo30Path, EDTEM& dtem, wxString& dataPath, wxString& mapQTHPath)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString srtmKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_SRTM_PATH;
	wxString globeKey   = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_GLOBE_PATH;
	wxString gtopo30Key = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_GTOPO30_PATH;
	wxString dtemKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_DTEM;
	wxString dataKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_DATA_PATH;
	wxString mapQTHKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_MAP_QTH_PATH;

	long fmt;
	profile->Read(srtmKey,    &srtmPath,    DEFAULT_SRTM_PATH);
	profile->Read(globeKey,   &globePath,   DEFAULT_GLOBE_PATH);
	profile->Read(gtopo30Key, &gtopo30Path, DEFAULT_GTOPO30_PATH);
	profile->Read(dtemKey,    &fmt,         DEFAULT_DTEM);
	profile->Read(dataKey,    &dataPath,    DEFAULT_DATA_PATH);
	profile->Read(mapQTHKey,  &mapQTHPath,  DEFAULT_MAP_QTH_PATH);

	dtem = EDTEM(fmt);
}

void CApp::setGeneralPreferences(const wxString& srtmPath, const wxString& globePath, const wxString& gtopo30Path, EDTEM dtem, const wxString& dataPath, const wxString& mapQTHPath)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString srtmKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_SRTM_PATH;
	wxString globeKey   = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_GLOBE_PATH;
	wxString gtopo30Key = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_GTOPO30_PATH;
	wxString dtemKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_DTEM;
	wxString dataKey    = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_DATA_PATH;
	wxString mapQTHKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_MAP_QTH_PATH;

	profile->Write(srtmKey,    srtmPath);
	profile->Write(globeKey,   globePath);
	profile->Write(gtopo30Key, gtopo30Path);
	profile->Write(dtemKey,    long(dtem));
	profile->Write(dataKey,    dataPath);
	profile->Write(mapQTHKey,  mapQTHPath);
	profile->Flush();
}

void CApp::get3DMapPreferences(int& mapSize)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString sizeKey = wxT("/") + SECTION_3D_MAP + wxT("/") + KEY_MAP_SIZE;

	long size;
	profile->Read(sizeKey, &size, DEFAULT_MAP_SIZE);

	mapSize = int(size);
}

void CApp::set3DMapPreferences(int mapSize)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString sizeKey = wxT("/") + SECTION_3D_MAP + wxT("/") + KEY_MAP_SIZE;

	profile->Write(sizeKey, long(mapSize));
	profile->Flush();
}

void CApp::getCoveragePreferences(int& height, int& frequency, int& txPower, int& txAntGain, int& rxPower, int& rxAntGain, int& polarisation, int& climate, int& ground, int& confidence)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString hgtKey     = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_HEIGHT;
	wxString freqKey    = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_FREQUENCY;
	wxString txPwrKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_TX_POWER;
	wxString txAntKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_TX_ANT_GAIN;
	wxString rxPwrKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_RX_MIN_SIG;
	wxString rxAntKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_RX_ANT_GAIN;
	wxString polKey     = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_POLARISATION;
	wxString climateKey = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_CLIMATE;
	wxString groundKey  = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_GROUND;
	wxString confiKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_CONFIDENCE;

	long hgt, freq, txPwr, txAnt, rxPwr, rxAnt, pol, clim, grnd, conf;
	profile->Read(hgtKey,     &hgt,         DEFAULT_DHEIGHT);
	profile->Read(freqKey,    &freq,        DEFAULT_FREQUENCY);
	profile->Read(txPwrKey,   &txPwr,       DEFAULT_TX_POWER);
	profile->Read(txAntKey,   &txAnt,       DEFAULT_TX_ANT_GAIN);
	profile->Read(rxPwrKey,   &rxPwr,       DEFAULT_RX_MIN_SIG);
	profile->Read(rxAntKey,   &rxAnt,       DEFAULT_RX_ANT_GAIN);
	profile->Read(polKey,     &pol,         DEFAULT_POLARISATION);
	profile->Read(climateKey, &clim,        DEFAULT_CLIMATE);
	profile->Read(groundKey,  &grnd,        DEFAULT_GROUND);
	profile->Read(confiKey,   &conf,        DEFAULT_CONFIDENCE);

	height       = int(hgt);
	frequency    = int(freq);
	txPower      = int(txPwr);
	txAntGain    = int(txAnt);
	rxPower      = int(rxPwr);
	rxAntGain    = int(rxAnt);
	polarisation = int(pol);
	climate      = int(clim);
	ground       = int(grnd);
	confidence   = int(conf);
}

void CApp::setCoveragePreferences(int height, int frequency, int txPower, int txAntGain, int rxPower, int rxAntGain, int polarisation, int climate, int ground, int confidence)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString hgtKey     = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_HEIGHT;
	wxString freqKey    = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_FREQUENCY;
	wxString txPwrKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_TX_POWER;
	wxString txAntKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_TX_ANT_GAIN;
	wxString rxPwrKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_RX_MIN_SIG;
	wxString rxAntKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_RX_ANT_GAIN;
	wxString polKey     = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_POLARISATION;
	wxString climateKey = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_CLIMATE;
	wxString groundKey  = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_GROUND;
	wxString confiKey   = wxT("/") + SECTION_COVERAGE + wxT("/") + KEY_CONFIDENCE;

	profile->Write(hgtKey,     long(height));
	profile->Write(freqKey,    long(frequency));
	profile->Write(txPwrKey,   long(txPower));
	profile->Write(txAntKey,   long(txAntGain));
	profile->Write(rxPwrKey,   long(rxPower));
	profile->Write(rxAntKey,   long(rxAntGain));
	profile->Write(polKey,     long(polarisation));
	profile->Write(climateKey, long(climate));
	profile->Write(groundKey,  long(ground));
	profile->Write(confiKey,   long(confidence));
	profile->Flush();
}

void CApp::getHorizonPreferences(int& distance, double& minElevation, double& maxElevation, double& k)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString distKey = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_DISTANCE;
	wxString minKey  = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_MIN_ELE;
	wxString maxKey  = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_MAX_ELE;
	wxString kKey    = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_K_VALUE;

	long dist;
	profile->Read(distKey, &dist,         DEFAULT_DISTANCE);
	profile->Read(minKey,  &minElevation, DEFAULT_MIN_ELE);
	profile->Read(maxKey,  &maxElevation, DEFAULT_MAX_ELE);
	profile->Read(kKey,    &k,            DEFAULT_K_VALUE);

	distance = int(dist);
}

void CApp::setHorizonPreferences(int distance, double minElevation, double maxElevation, double k)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString distKey = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_DISTANCE;
	wxString minKey  = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_MIN_ELE;
	wxString maxKey  = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_MAX_ELE;
	wxString kKey    = wxT("/") + SECTION_HORIZON + wxT("/") + KEY_K_VALUE;

	profile->Write(distKey, long(distance));
	profile->Write(minKey,  minElevation);
	profile->Write(maxKey,  maxElevation);
	profile->Write(kKey,    k);
	profile->Flush();
}

void CApp::getMapPreferences(EFormat& format, int& resolution, EType& type, bool& towns, bool& sites, int& textSize)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString formatKey = wxT("/") + SECTION_MAP + wxT("/") + KEY_FORMAT;
	wxString pixelKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_RESOLUTION;
	wxString typeKey   = wxT("/") + SECTION_MAP + wxT("/") + KEY_TYPE;
	wxString townsKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_TOWNS;
	wxString sitesKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_SITES;
	wxString textKey   = wxT("/") + SECTION_MAP + wxT("/") + KEY_TEXT_SIZE;

	long fmt, res, typ, text;
	profile->Read(formatKey, &fmt,   DEFAULT_FORMAT);
	profile->Read(pixelKey,  &res,   DEFAULT_RESOLUTION);
	profile->Read(typeKey,   &typ,   DEFAULT_TYPE);
	profile->Read(townsKey,  &towns, DEFAULT_TOWNS);
	profile->Read(sitesKey,  &sites, DEFAULT_SITES);
	profile->Read(textKey,   &text,  DEFAULT_TEXT_SIZE);

	format     = EFormat(fmt);
	resolution = int(res);
	type       = EType(typ);
	textSize   = int(text);
}

void CApp::setMapPreferences(EFormat format, int resolution, EType type, bool towns, bool sites, int textSize)
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString formatKey = wxT("/") + SECTION_MAP + wxT("/") + KEY_FORMAT;
	wxString pixelKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_RESOLUTION;
	wxString typeKey   = wxT("/") + SECTION_MAP + wxT("/") + KEY_TYPE;
	wxString townsKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_TOWNS;
	wxString sitesKey  = wxT("/") + SECTION_MAP + wxT("/") + KEY_SITES;
	wxString textKey   = wxT("/") + SECTION_MAP + wxT("/") + KEY_TEXT_SIZE;

	profile->Write(formatKey, long(format));
	profile->Write(pixelKey,  long(resolution));
	profile->Write(typeKey,   long(type));
	profile->Write(townsKey,  towns);
	profile->Write(sitesKey,  sites);
	profile->Write(textKey,   long(textSize));
	profile->Flush();
}

void CApp::getMap(wxString& fileName, CLocation& centre, int& resolution, EType& type, bool& towns, bool& sites)
{
	fileName   = m_fileName;
	centre     = m_centre;
	resolution = m_resolution;
	type       = m_type;
	towns      = m_towns;
	sites      = m_sites;
}

void CApp::setMap(const wxString& fileName, const CLocation& centre, int resolution, EType type, bool towns, bool sites)
{
	m_fileName   = fileName;
	m_centre     = centre;
	m_resolution = resolution;
	m_type       = type;
	m_towns      = towns;
	m_sites      = sites;
}

CDEMCache* CApp::getDEMCache()
{
	return m_cache;
}

CPoint* CApp::getPoint(int n)
{
	wxASSERT(n >= 0 && n < MAX_POINT_COUNT);

	return m_points[n];
}

int CApp::setPoint(CPoint* location)
{
	for (int i = 0; i < MAX_POINT_COUNT; i++) {
		if (m_points[i] == NULL) {
			m_points[i] = location;
			return i;
		}
	}

	return -1;
}

int CApp::countPoints() const
{
	int n = 0;

	for (int i = 0; i < MAX_POINT_COUNT; i++) {
		if (m_points[i] != NULL)
			n++;
	}

	return n;
}

bool CApp::removePoint(int n)
{
	wxASSERT(n >= 0 && n < MAX_POINT_COUNT);

	if (m_points[n] == NULL)
		return false;

	delete m_points[n];
	m_points[n] = NULL;

	return true;
}

void CApp::clearPoints()
{
	for (int i = 0; i < MAX_POINT_COUNT; i++) {
		if (m_points[i] != NULL) {
			delete m_points[i];
			m_points[i] = NULL;
		}
	}
}

void CApp::clearPoints(double minLat, double minLon, double maxLat, double maxLon)
{
	if (minLat > maxLat) {
		double temp = maxLat;
		maxLat = minLat;
		minLat = temp;
	}

	if (minLon > maxLon) {
		double temp = maxLon;
		maxLon = minLon;
		minLon = temp;
	}

	for (int i = 0; i < MAX_POINT_COUNT; i++) {
		if (m_points[i] != NULL) {
			CLocation location = m_points[i]->getLocation();

			double latitude, longitude;
			location.getLatLon(latitude, longitude);

			if (latitude  < minLat || latitude  > maxLat ||
			    longitude < minLon || longitude > maxLon) {
				delete m_points[i];
				m_points[i] = NULL;
			}
		}
	}
}

