/*
 *   Copyright (C) 2003-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Coverage.h"

#include <wx/progdlg.h>

const double SURFREF = 301.0;

CCoverage::CCoverage(int* map, int resolution, double ulLat, double ulLon, double brLat, double brLon) :
m_map(map),
m_resolution(resolution),
m_ulLat(ulLat),
m_ulLon(ulLon),
m_brLat(brLat),
m_brLon(brLon),
m_stepLat(0.0),
m_stepLon(0.0),
m_frequency(0),
m_polarisation(0),
m_climate(0),
m_ground(0),
m_confidence(0),
m_maxPathLoss(0.0)
{
	wxASSERT(m_map != NULL);

	m_stepLat = (m_ulLat - m_brLat) / double(GRAPH_HEIGHT);
	m_stepLon = (m_brLon - m_ulLon) / double(GRAPH_WIDTH);

}

CCoverage::~CCoverage()
{
}

void CCoverage::setITM(int frequency, int polarisation, int climate, int ground, int confidence)
{
	wxASSERT(frequency >= 20 && frequency <= 20000);
	wxASSERT(polarisation == 0 || polarisation == 1);
	wxASSERT(climate >= 1 && climate <= 7);
	wxASSERT(ground >= 0 && ground <= 8);
	wxASSERT(confidence >= 1 && confidence <= 99);

	m_frequency    = frequency;
	m_polarisation = polarisation;
	m_climate      = climate;
	m_ground       = ground;
	m_confidence   = confidence;
}

void CCoverage::setRFParams(int txPower, int txAntGain, int rxPower, int rxAntGain)
{
	m_maxPathLoss = -double(rxPower - rxAntGain - txAntGain - txPower);
}

void CCoverage::calculate(int* coverage, const CLocation& location, int startHeight, int endHeight)
{
	wxASSERT(coverage != NULL);

	wxProgressDialog dialog(wxT("Working..."), wxT("Calculating Coverage"), 360, this);

	for (int i = 0; i < (GRAPH_WIDTH * GRAPH_HEIGHT); i++)
		coverage[i] = VOID_DATA;

	int aveHeight = getAveHeight();

	CITM itm(aveHeight, SURFREF, m_frequency, m_climate, m_ground, m_polarisation, m_confidence);

	int maxPixels = ::max(GRAPH_WIDTH, GRAPH_HEIGHT);

	double maxDist = maxPixels * double(m_resolution) / 1000.0;

	double* elev = new double[maxPixels + 2];

	CPath path(location);

	for (double az = 0.0; az < 360.0; az += 0.25) {
		dialog.Update(int(az));

		path.setDestAzDist(az, maxDist);

		getLOS(coverage, itm, elev, path, startHeight, endHeight);
	}

	delete[] elev;
}

void CCoverage::getLOS(int* coverage, CITM& itm, double* elev, const CPath& path, int startHeight, int endHeight)
{
	wxASSERT(elev != NULL);
	wxASSERT(m_map != NULL);
	wxASSERT(coverage != NULL);

	double stepDeg = ::min(m_stepLat, m_stepLon);
	double step = R * RAD(stepDeg);

	double distance = path.getDistance();

	int n = 0;

	int lastX = 0;
	int lastY = 0;

	for (double dist = step; dist < distance; dist += step, n++) {
		double lat, lon;
		path.getPath(dist, lat, lon);

		int x = longitudeToX(lon);
		int y = latitudeToY(lat);

		if (x < 0 || x >= GRAPH_WIDTH || y < 0 || y >= GRAPH_HEIGHT)
			return;

		int index = y * GRAPH_WIDTH + x;

		int height = m_map[index];
		if (height == VOID_DATA)
			height = 0;

		elev[n + 2] = double(height);

		elev[0] = double(n);
		elev[1] = step * 1000.0;

		if (n > 0 && (x != lastX || y != lastY) && coverage[index] == VOID_DATA) {
			double dbloss;
			wxString info;
			itm.calculate(elev, startHeight, endHeight, dbloss, info);

			int err = itm.getErr();

			if (err == 0 || err == 1 || err == 2 || err == 3)
				coverage[index] = int(m_maxPathLoss - dbloss + 0.5);
		}

		lastX = x;
		lastY = y;
	}
}

double CCoverage::xToLongitude(int x) const
{
	wxASSERT(x >= 0 || x < GRAPH_WIDTH);

	double lon = m_ulLon + double(x) * m_stepLon;

	if (lon > 180.0)
		lon -= 360.0;

	if (lon < -180.0)
		lon += 360.0;

	return lon;
}

double CCoverage::yToLatitude(int y) const
{
	wxASSERT(y >= 0 || y < GRAPH_HEIGHT);

	return m_ulLat - double(y) * m_stepLat;
}

int CCoverage::longitudeToX(double lon) const
{
	int x = int((lon - m_ulLon) / m_stepLon);

	return (x < 0 || x >= GRAPH_WIDTH) ? -1 : x;
}

int CCoverage::latitudeToY(double lat) const
{
	int y = int((m_ulLat - lat) / m_stepLat);

	return (y < 0 || y >= GRAPH_HEIGHT) ? -1 : y;
}

int CCoverage::getAveHeight() const
{
	int n = 0;
	int total = 0;

	for (int x = 0; x < GRAPH_WIDTH; x += GRAPH_WIDTH / 8) {
		for (int y = 0; y < GRAPH_HEIGHT; y += GRAPH_HEIGHT / 8) {
			int index = y * GRAPH_WIDTH + x;

			int height = m_map[index];
			if (height == VOID_DATA)
				height = 0;

			total += height;
			n++;
		}
	}

	return total / n;
}
