/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <wx/tokenzr.h>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/log.h>

#include "GTOPO30.h"
#include "Defs.h"

const struct {
	wxChar* name;
	double  minLatitude;
	double  maxLatitude;
	double  minLongitude;
	double  maxLongitude;
} Maps[] = {
	{wxT("W180N90"),  40.0,  90.0, -180.0, -140.0},
	{wxT("W140N90"),  40.0,  90.0, -140.0, -100.0},
	{wxT("W100N90"),  40.0,  90.0, -100.0,  -60.0},
	{wxT("W060N90"),  40.0,  90.0,  -60.0,  -20.0},
	{wxT("W020N90"),  40.0,  90.0,  -20.0,   20.0},
	{wxT("E020N90"),  40.0,  90.0,   20.0,   60.0},
	{wxT("E060N90"),  40.0,  90.0,   60.0,  100.0},
	{wxT("E100N90"),  40.0,  90.0,  100.0,  140.0},
	{wxT("E140N90"),  40.0,  90.0,  140.0,  180.0},
	{wxT("W180N40"), -10.0,  40.0, -180.0, -140.0},
	{wxT("W140N40"), -10.0,  40.0, -140.0, -100.0},
	{wxT("W100N40"), -10.0,  40.0, -100.0,  -60.0},
	{wxT("W060N40"), -10.0,  40.0,  -60.0,  -20.0},
	{wxT("W020N40"), -10.0,  40.0,  -20.0,   20.0},
	{wxT("E020N40"), -10.0,  40.0,   20.0,   60.0},
	{wxT("E060N40"), -10.0,  40.0,   60.0,  100.0},
	{wxT("E100N40"), -10.0,  40.0,  100.0,  140.0},
	{wxT("E140N40"), -10.0,  40.0,  140.0,  180.0},
	{wxT("W180S10"), -60.0, -10.0, -180.0, -140.0},
	{wxT("W140S10"), -60.0, -10.0, -140.0, -100.0},
	{wxT("W100S10"), -60.0, -10.0, -100.0,  -60.0},
	{wxT("W060S10"), -60.0, -10.0,  -60.0,  -20.0},
	{wxT("W020S10"), -60.0, -10.0,  -20.0,   20.0},
	{wxT("E020S10"), -60.0, -10.0,   20.0,   60.0},
	{wxT("E060S10"), -60.0, -10.0,   60.0,  100.0},
	{wxT("E100S10"), -60.0, -10.0,  100.0,  140.0},
	{wxT("E140S10"), -60.0, -10.0,  140.0,  180.0},
};

const int numMaps = sizeof(Maps) / sizeof(Maps[0]);

CGTOPO30::CGTOPO30(const wxString& path, double latitude, double longitude) :
CDEM(),
m_index(-1),
m_rowBytes(0),
m_rows(0),
m_cols(0),
m_noData(0),
m_ulX(0.0),
m_ulY(0.0),
m_dimX(0.0),
m_dimY(0.0),
m_file(NULL)
{
	m_index = findFile(latitude, longitude);

	if (m_index == -1)
		return;

	wxString name = Maps[m_index].name;

	wxFileName headerFileName(path, name, wxT("HDR"));

	if (!headerFileName.IsOk())
		return;

	bool ret = readHeader(headerFileName.GetFullPath());

	if (ret) {
		wxFileName dataFileName(path, name, wxT("DEM"));

		if (!dataFileName.IsOk())
			return;

		m_file = new wxFFile(dataFileName.GetFullPath().c_str(), "r");

		ret = m_file->IsOpened();

		if (!ret) {
			wxString message = wxT("Cannot open the GTOPO30 data file: ") + dataFileName.GetFullPath();
			::wxLogFatalError(message);
			exit(1);
		}
	}
}

CGTOPO30::~CGTOPO30()
{
	if (m_file != NULL)
		delete m_file;
}

int CGTOPO30::getHeight(double latitude, double longitude)
{
	wxASSERT(latitude >= Maps[m_index].minLatitude);
	wxASSERT(latitude < Maps[m_index].maxLatitude);
	wxASSERT(longitude >= Maps[m_index].minLongitude);
	wxASSERT(longitude < Maps[m_index].maxLongitude);

	if (m_file == NULL)
		return VOID_DATA;

	int row = int((m_ulY - latitude) / m_dimY);
	int col = int((longitude - m_ulX) / m_dimX);

	wxASSERT(row >= 0 && row < m_rows);
	wxASSERT(col >= 0 && col < m_cols);

	return readHeight(row, col);
}

void CGTOPO30::setHeight(int height, double latitude, double longitude)
{
}

bool CGTOPO30::isInData(double latitude, double longitude) const
{
	return (latitude  >= Maps[m_index].minLatitude  &&
		latitude  <  Maps[m_index].maxLatitude  &&
		longitude >= Maps[m_index].minLongitude &&
		longitude <  Maps[m_index].maxLongitude);
}

bool CGTOPO30::dataExists() const
{
	return m_file != NULL;
}

double CGTOPO30::getMinLatitude() const
{
	return Maps[m_index].minLatitude;
}

double CGTOPO30::getMaxLatitude() const
{
	return Maps[m_index].maxLatitude;
}

double CGTOPO30::getMinLongitude() const
{
	return Maps[m_index].minLongitude;
}

double CGTOPO30::getMaxLongitude() const
{
	return Maps[m_index].maxLongitude;
}

int CGTOPO30::findFile(double latitude, double longitude) const
{
	for (int i = 0; i < numMaps; i++) {
		if (latitude >= Maps[i].minLatitude &&
		    latitude < Maps[i].maxLatitude &&
		    longitude >= Maps[i].minLongitude &&
		    longitude < Maps[i].maxLongitude)
			return i;
	}

	return -1;
}

bool CGTOPO30::readHeader(const wxString& fileName)
{
	wxTextFile file;

	if (!file.Open(fileName))
		return false;

	int n = file.GetLineCount();

	for (int line = 0; line < n; line++) {
		wxString str = file.GetLine(line);

		processHeader(str);
	}

	file.Close();

	return true;
}

void CGTOPO30::processHeader(const wxString& line)
{
	wxStringTokenizer token(line);

	wxString name = token.GetNextToken();

	if (name.IsSameAs(wxT("NROWS")))
		token.GetNextToken().ToLong(&m_rows);
	else if (name.IsSameAs(wxT("NCOLS")))
		token.GetNextToken().ToLong(&m_cols);
	else if (name.IsSameAs(wxT("TOTALROWBYTES")))
		token.GetNextToken().ToLong(&m_rowBytes);
	else if (name.IsSameAs(wxT("NODATA")))
		token.GetNextToken().ToLong(&m_noData);
	else if (name.IsSameAs(wxT("ULXMAP")))
		token.GetNextToken().ToDouble(&m_ulX);
	else if (name.IsSameAs(wxT("ULYMAP")))
		token.GetNextToken().ToDouble(&m_ulY);
	else if (name.IsSameAs(wxT("XDIM")))
		token.GetNextToken().ToDouble(&m_dimX);
	else if (name.IsSameAs(wxT("YDIM")))
		token.GetNextToken().ToDouble(&m_dimY);
}

int CGTOPO30::readHeight(int row, int col) const
{
	if (row < 0 || row >= m_rows ||
	    col < 0 || col >= m_cols)
		return VOID_DATA;

	long offset = row * m_rowBytes + col * sizeof(wxInt16);

	m_file->Seek(offset);

	wxInt16 data;
	m_file->Read(&data, sizeof(wxInt16));

	int height = wxINT16_SWAP_ON_LE(data);

	if (height == m_noData)
		return VOID_DATA;

	return height;
}
