/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <wx/tokenzr.h>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/log.h>

#include "Globe.h"
#include "Defs.h"

const struct {
	wxChar  prefix;
	double  minLatitude;
	double  maxLatitude;
	double  minLongitude;
	double  maxLongitude;
} Maps[] = {
	{wxT('a'),  50.0,  90.0, -180.0, -90.0},
	{wxT('b'),  50.0,  90.0,  -90.0,   0.0},
	{wxT('c'),  50.0,  90.0,    0.0,  90.0},
	{wxT('d'),  50.0,  90.0,   90.0, 180.0},
	{wxT('e'),   0.0,  50.0, -180.0, -90.0},
	{wxT('f'),   0.0,  50.0,  -90.0,   0.0},
	{wxT('g'),   0.0,  50.0,    0.0,  90.0},
	{wxT('h'),   0.0,  50.0,   90.0, 180.0},
	{wxT('i'), -50.0,   0.0, -180.0, -90.0},
	{wxT('j'), -50.0,   0.0,  -90.0,   0.0},
	{wxT('k'), -50.0,   0.0,    0.0,  90.0},
	{wxT('l'), -50.0,   0.0,   90.0, 180.0},
	{wxT('m'), -90.0, -50.0, -180.0, -90.0},
	{wxT('n'), -90.0, -50.0,  -90.0,   0.0},
	{wxT('o'), -90.0, -50.0,    0.0,  90.0},
	{wxT('p'), -90.0, -50.0,   90.0, 180.0},
};

const int numMaps = sizeof(Maps) / sizeof(Maps[0]);

const double globeCellSize = 1.0 / 120.0;

CGlobe::CGlobe(const wxString& path, EDTEM version, double latitude, double longitude) :
CDEM(),
m_index(-1),
m_file(NULL)
{
	m_index = findFile(latitude, longitude);

	if (m_index == -1)
		return;

	wxChar prefix = Maps[m_index].prefix;
	wxString name;

	switch (version) {
		case DTEM_GLOBE_V1:
			name.Printf(wxT("%c10g"), prefix);
			break;
		case DTEM_GLOBE_V2:
			name.Printf(wxT("%c20g"), prefix);
			break;
		default:
			wxASSERT(false);
			m_index = -1;
			return;
	}

	wxFileName fileName(path, name);

	if (!fileName.IsOk())
		return;

	m_file = new wxFFile(fileName.GetFullPath().c_str(), "r");

	bool ret = m_file->IsOpened();

	if (!ret) {
		wxString message = wxT("Cannot open the Globe data file: ") + fileName.GetFullPath();
		::wxLogFatalError(message);
		exit(1);
	}
}

CGlobe::~CGlobe()
{
	if (m_file != NULL)
		delete m_file;
}

int CGlobe::getHeight(double latitude, double longitude)
{
	wxASSERT(latitude >= Maps[m_index].minLatitude);
	wxASSERT(latitude < Maps[m_index].maxLatitude);
	wxASSERT(longitude >= Maps[m_index].minLongitude);
	wxASSERT(longitude < Maps[m_index].maxLongitude);

	int row = int((Maps[m_index].maxLatitude - latitude) / globeCellSize);
	int col = int((longitude - Maps[m_index].minLongitude) / globeCellSize);

	return readHeight(row, col);
}

void CGlobe::setHeight(int height, double latitude, double longitude)
{
}

bool CGlobe::isInData(double latitude, double longitude) const
{
	return (latitude  >= Maps[m_index].minLatitude  &&
		latitude  <  Maps[m_index].maxLatitude  &&
		longitude >= Maps[m_index].minLongitude &&
		longitude <  Maps[m_index].maxLongitude);
}

bool CGlobe::dataExists() const
{
	return m_file != NULL;
}

double CGlobe::getMinLatitude() const
{
	return Maps[m_index].minLatitude;
}

double CGlobe::getMaxLatitude() const
{
	return Maps[m_index].maxLatitude;
}

double CGlobe::getMinLongitude() const
{
	return Maps[m_index].minLongitude;
}

double CGlobe::getMaxLongitude() const
{
	return Maps[m_index].maxLongitude;
}

int CGlobe::findFile(double latitude, double longitude) const
{
	for (int i = 0; i < numMaps; i++) {
		if (latitude >= Maps[i].minLatitude &&
		    latitude < Maps[i].maxLatitude &&
		    longitude >= Maps[i].minLongitude &&
		    longitude < Maps[i].maxLongitude)
			return i;
	}

	return -1;
}

int CGlobe::readHeight(int row, int col) const
{
	if (row < 0 || col < 0)
		return VOID_DATA;

	long offset = (row * 10800 + col) * sizeof(wxInt16);

	m_file->Seek(offset);

	wxInt16 data;
	m_file->Read(&data, sizeof(wxInt16));

	int height = wxINT16_SWAP_ON_BE(data);

	if (height == -500)
		return VOID_DATA;

	return height;
}
