/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Horizon.h"
#include "Path.h"
#include "Defs.h"
#include "App.h"

const int numColours = 11;

BEGIN_EVENT_TABLE(CHorizon, wxPanel)
	EVT_PAINT(CHorizon::onPaint)
END_EVENT_TABLE()

CHorizon::CHorizon(wxWindow* parent, const CPoint& point, int distance, int startAngle, int endAngle, double minEle, double maxEle, double k) :
wxPanel(parent, -1, wxDefaultPosition, wxSize(GRAPH_WIDTH, GRAPH_HEIGHT)),
m_point(point),
m_height(0),
m_k(k),
m_distance(distance),
m_startAngle(startAngle),
m_endAngle(endAngle),
m_minEle(minEle),
m_maxEle(maxEle),
m_bitmap(NULL),
m_palette(NULL)
{
	CDEMCache* cache = ::wxGetApp().getDEMCache();

	wxASSERT(cache != NULL);

	m_bitmap  = new wxBitmap(GRAPH_WIDTH, GRAPH_HEIGHT);
	m_palette = new wxColour*[numColours];

	CLocation location = m_point.getLocation();

	double lat = location.getLatitude();
	double lon = location.getLongitude();

	if (m_point.getASL() != VOID_DATA) {
		m_height = m_point.getASL();
	} else {
		m_height = cache->getHeight(lat, lon);

		if (m_height == VOID_DATA)
			m_height = 0;
	}

	m_height += m_point.getHeight();

	createPalette();

	drawMap();
	drawScale();
	drawInfo();
}

CHorizon::~CHorizon()
{
	delete m_bitmap;

	for (int i = 0; i < numColours; i++)
		delete m_palette[i];
	delete[] m_palette;
}

void CHorizon::createPalette()
{
	for (int i = 0; i < numColours; i++) {
		int r = i * 16 + 79;
		int g = 243 - i * 12;
		int b = 235 - i * 20;

		m_palette[i] = new wxColour(r, g, b);
	}
}

void CHorizon::drawMap()
{
	CDEMCache* cache = ::wxGetApp().getDEMCache();

	wxASSERT(cache != NULL);

	// Flood the graph area with black to start with
	wxMemoryDC dc;
	dc.SelectObject(*m_bitmap);
	dc.BeginDrawing();
	dc.SetBackground(*wxBLACK_BRUSH);
	dc.Clear();

	dc.SetPen(*wxWHITE_PEN);

	int x1 = azimuthToX(double(m_startAngle));
	int x2 = azimuthToX(double(m_endAngle));
	int y1 = elevationToY(m_minEle);
	int y2 = elevationToY(m_maxEle);
	
	dc.DrawLine(x1, y1, x2, y1);
	dc.DrawLine(x2, y1, x2, y2);

	dc.SetTextForeground(*wxWHITE);
	for (int azimuth = m_startAngle; azimuth <= m_endAngle; azimuth += 30) {
		int x = azimuthToX(double(azimuth));

		dc.DrawLine(x, y1, x, y1 + 10);

		wxString text;
		text.Printf(wxT("%d"), azimuth);

		wxCoord width, height;
		dc.GetTextExtent(text, &width, &height);

		dc.DrawText(text, x - width / 2, y1 + 15);
	}

	for (double elevation = m_minEle; elevation <= m_maxEle; elevation += 0.5) {
		int y = elevationToY(elevation);

		dc.DrawLine(x2, y, x2 + 10, y);

		wxString text;
		text.Printf(wxT("%1.2f"), elevation);
		dc.DrawText(text, x2 + 15, y - 5);
	}

	dc.DrawText(wxT("Bearing"), x2 + 70, y1 + 15);
	dc.DrawText(wxT("Horizon"), x2 + 70, y1 - 55);
	dc.DrawText(wxT("angle"),   x2 + 70, y1 - 40);

	wxPen** pens = new wxPen*[numColours];
	for (int i = 0; i < numColours; i++)
		pens[i] = new wxPen(*m_palette[i], 1, wxSOLID);

	double re = 6371.4 * m_k;

	double azIncr = double(m_endAngle - m_startAngle) / 540.0;

	CPath path(m_point.getLocation());

	for (double azimuth = double(m_startAngle); azimuth < double(m_endAngle); azimuth += azIncr) {
		int x = azimuthToX(azimuth);

		path.setDestAzDist(azimuth, m_distance);

		double maxTheta = m_minEle;

		for (double along = 0.5; along < double(m_distance); along += 0.5) {
			int colour = (int(along) * 9) / m_distance + 2;

			double lat, lon;
			path.getPath(along, lat, lon);

			int height = cache->getHeight(lat, lon, true);

			if (height != VOID_DATA) {
				double theta = DEG(double(height - m_height) / 1000.0 / along - along / 2.0 / re);

				if (theta < m_minEle)
					theta = m_minEle;

				if (theta > maxTheta) {
					maxTheta = theta;
					int y = elevationToY(theta);

					dc.SetPen(*pens[colour]);
					dc.DrawPoint(x, y);
				}
			}
		}
	}

	for (int i = 0; i < numColours; i++)
		delete pens[i];
	delete[] pens;

	dc.EndDrawing();
}

void CHorizon::drawScale()
{
	wxMemoryDC dc;
	dc.SelectObject(*m_bitmap);
	dc.BeginDrawing();

	dc.SetTextForeground(*wxWHITE);
	dc.DrawText(wxT("Distance"), GRAPH_WIDTH - 70, 30);

	for (int i = 1; i < numColours; i++) {
		double distance = m_distance * (i - 1) / 9;

		wxString text;
		text.Printf(wxT("%.0f km"), distance);
		dc.DrawText(text, GRAPH_WIDTH - 80, 205 - i * 15);

		wxBrush brush(*m_palette[i], wxSOLID);
		dc.SetBrush(brush);
		dc.DrawRectangle(GRAPH_WIDTH - 30, 205 - i * 15, 15, 15);
	}

	dc.EndDrawing();	
}

void CHorizon::drawInfo()
{
	wxMemoryDC dc;
	dc.SelectObject(*m_bitmap);
	dc.BeginDrawing();

	dc.SetTextForeground(*wxWHITE);

	CLocation location = m_point.getLocation();

	EFormat format = location.getFormat();

	wxString text;
	text.Printf(wxT("Horizon view from %s"), location.getLocation(format).c_str());
	dc.DrawText(text, 20, 10);

	text.Printf(wxT("%dm ASL"), m_height);
	dc.DrawText(text, 20, 25);

	dc.EndDrawing();	
}

void CHorizon::onPaint(const wxPaintEvent& event)
{
	wxPaintDC dc(this);

	dc.BeginDrawing();
	dc.DrawBitmap(*m_bitmap, 0, 0, false);
	dc.EndDrawing();
}

int CHorizon::azimuthToX(double azimuth) const
{
	wxASSERT(azimuth >= 0.0 && azimuth <= 360.0);

	double scale = 540.0 / double(m_endAngle - m_startAngle);

	return int((azimuth - double(m_startAngle)) * scale) + 30;
}

int CHorizon::elevationToY(double elevation) const
{
	if (elevation < m_minEle)
		elevation = m_minEle;

	if (elevation > m_maxEle)
		elevation = m_maxEle;

	double scale = 450.0 / (m_maxEle - m_minEle);

	return GRAPH_HEIGHT - 40 - int((elevation - m_minEle) * scale);
}
