/****************************************************************************
*                                                                           *
* This file was obtained from ftp://flattop.its.bldrdoc.gov/itm/ITMDLL.cpp  *
* on Jan. 10, 2004 and is public domain.  It was modified by J. D. McDonald *
* to remove Microsoft Windows dll-isms and to correct one case where a      *
* compiler found an ambguity in overloaded calls.  It was further modified  *
* by John A. Magliacane to remove unused variables, unneeded #includes.     *
*                                                                           *
****************************************************************************/

// *************************************
// C++ routines for this program are taken from
// a translation of the FORTRAN code written by
// U.S. Department of Commerce NTIA/ITS
// Institute for Telecommunication Sciences
// *****************
// Irregular Terrain Model (ITM) (Longley-Rice)
// *************************************

#include "ITM.h"

CITM::CITM(int zsys, double surfref, int f, int climate, int ground, int polarisation, int confidence) :
m_aref(0.0),
m_dist(0.0),
m_hg(),
m_wn(0.0),
m_dh(0.0),
m_ens(0.0),
m_gme(0.0),
m_zgnd(0.0, 0.0),
m_he(),
m_dl(),
m_the(),
m_kwx(0),
m_sgc(0.0),
m_klim(climate),
m_dlsa(0.0),
m_dx(0.0),
m_ael(0.0),
m_ak1(0.0),
m_ak2(0.0),
m_aed(0.0),
m_emd(0.0),
m_aes(0.0),
m_ems(0.0),
m_dls(),
m_dla(0.0),
m_tha(0.0),
m_wd1(0.0),
m_xd1(0.0),
m_afo(0.0),
m_qk(0.0),
m_aht(0.0),
m_xht(0.0),
m_ad(0.0),
m_rr(0.0),
m_etq(0.0),
m_h0s(0.0),
m_wls(0.0),
m_zc(0.0),
m_f(f)
{
	wxASSERT(f >= 20 && f <= 20000);
	wxASSERT(climate >= 0 && climate < 7);
	wxASSERT(ground >= 0 && ground < 8);
	wxASSERT(polarisation == 0 || polarisation == 1);
	wxASSERT(confidence >= 1 && confidence <= 99);

	const double dielectric[]   = {80.0,  25.0,  80.0,  12.0,  15.0,  15.0,  13.0,  5.0,   4.0};
	const double conductivity[] = {5.000, 0.020, 0.010, 0.007, 0.005, 0.005, 0.002, 0.001, 0.001};

	m_hg[0] = 0.0;
	m_hg[1] = 0.0;

	m_he[0] = 0.0;
	m_he[1] = 0.0;

	m_dl[0] = 0.0;
	m_dl[1] = 0.0;

	m_the[0] = 0.0;
	m_the[1] = 0.0;

	m_dls[0] = 0.0;
	m_dls[1] = 0.0;

	double eps  = dielectric[ground];
	double sgm  = conductivity[ground];

	qlrps(zsys, surfref, polarisation, eps, sgm);

	m_zc = qerfi(double(confidence) / 100.0);
}

CITM::~CITM()
{
}

/*
 * This performs the FORTRAN DIM function. Result is x - y
 * if x is greater than y; otherwise result is 0.0
 */
double CITM::FORTRAN_DIM(double x, double y) const
{
	if (x > y)
		return x - y;
	else
		return 0.0;
}

/*
 * Calculate the knife edge attenuation.
 *
 * Checked againt ITM.pdf, 25.2.2004
 */
double CITM::knifeEdgeAttenuation(double v2) const
{
	if (v2 < 5.76)
		return 6.02 + 9.11 * ::sqrt(v2) - 1.27 * v2;
	else
		return 12.953 + 4.343 * ::log(v2);
}

/*
 * Calculate the height gain over a smooth spherical earth.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::heightGain(double x, double pk) const
{
	double fht;

	if (x < 200.0) {
		double w = -::log(pk);

		if (pk < 1.0e-5 || x * w * w * w > 5495.0) {
			fht = -117.0;

			if (x > 1.0)
				fht += 17.372 * ::log(x);
		} else {
			fht = 2.5e-5 * x * x / pk - 8.686 * w - 15.0;
		}
	} else {
		fht = 0.05751 * x - 4.343 * ::log(x);

		if (x < 2000.0) {
			double w = 0.0134 * x * ::exp(-0.005 * x);

			fht = (1.0 - w) * fht + w * (17.372 * ::log(x) - 117.0);
		}
	}

	return fht;
}

/*
 * Calculate the H0 function for scatter fields.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::h0f(double r, double et) const
{
	const double a[5] = {25.0, 80.0, 177.0, 395.0, 705.0};
	const double b[5] = {24.0, 45.0,  68.0,  80.0, 105.0};

	int it = int(et);
	double q;

	if (it <= 0) {
		it = 1;
		q  = 0.0;
	} else if (it >= 5) {
		it = 5;
		q  = 0.0;
	} else {
		q = et - double(it);
	}

	double x = ::pow(1.0 / r, 2.0);

	double h0f = 4.343 * ::log((a[it - 1] * x + b[it - 1]) * x + 1.0);

	if (q != 0.0)
		h0f = (1.0 - q) * h0f + q * 4.343 * ::log((a[it] * x + b[it]) * x + 1.0);

	return h0f;
}

/*
 * Calculate the F(thetaD) function for scatter fields.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::ahd(double td) const
{
	const double a[3] = {   133.4,    104.6,     71.8};
	const double b[3] = {0.332e-3, 0.212e-3, 0.157e-3};
	const double c[3] = {  -4.343,   -1.086,    2.171};

	int i;
	if (td <= 10e3)
		i = 0;
	else if (td <= 70e3)
		i = 1;
	else
		i = 2;

	return a[i] + b[i] * td + c[i] * ::log(td);
}

/*
 * Calculate the diffraction attenuation.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::diffractionAttenuation(double d)
{
	double adiff;

	if (d == 0.0) {
		// Prepare initial diffraction constants - 11
		double q = m_hg[0] * m_hg[1];
		m_qk = m_he[0] * m_he[1] - q;

		q += 10.0;

		m_wd1 = ::sqrt(1.0 + m_qk / q);
		m_xd1 = m_dla + m_tha / m_gme;

		q  = (1.0 - 0.8 * ::exp(-m_dlsa / 50e3)) * m_dh;
		q *= 0.78 * ::exp(-::pow(q / 16.0, 0.25));

		m_afo = ::min(15.0, 2.171 * ::log(1.0 + 4.77e-4 * m_hg[0] * m_hg[1] * m_wn * q));
		m_qk  = 1.0 / abs(m_zgnd);

		m_aht = 20.0;
		m_xht = 0.0;

		for (int j = 0; j < 2; j++) {
			double a  = 0.5 * (m_dl[j] * m_dl[j]) / m_he[j];
			double wa = ::pow(a * m_wn, THIRD);
			double pk = m_qk / wa;

			q = (1.607 - pk) * 151.0 * wa * m_dl[j] / a;
			m_xht += q;
			m_aht += heightGain(q, pk);
		}

		adiff = 0.0;
	} else {
		// Compute diffraction attenuation - 12
		double th = m_tha + d * m_gme;
		double ds = d - m_dla;
		double q = 0.0795775 * m_wn * ds * th * th;

		adiff = knifeEdgeAttenuation(q * m_dl[0] / (ds + m_dl[0])) +
			knifeEdgeAttenuation(q * m_dl[1] / (ds + m_dl[1]));

		double a  = ds / th;
		double wa = ::pow(a * m_wn, THIRD);
		double pk = m_qk / wa;
		q = (1.607 - pk) * 151.0 * wa * th + m_xht;
		double ar = 0.05751 * q - 4.343 * ::log(q) - m_aht;
		q = (m_wd1 + m_xd1 / d) * ::min(((1.0 - 0.8 * ::exp(-d / 50e3)) * m_dh * m_wn), 6283.2);
		double wd = 25.1 / (25.1 + ::sqrt(q));

		adiff = ar * wd + (1.0 - wd) * adiff + m_afo;
	}

	return adiff;
}

/*
 * Calculate the scatter attenuation.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::scatterAttenuation(double d)
{
	double ascat;

	if (d == 0.0) {
		// Prepare initial scatter constants - 23
		m_ad = m_dl[0] - m_dl[1];
		m_rr = m_he[1] / m_he[0];

		if (m_ad < 0.0) {
			m_ad = -m_ad;
			m_rr = 1.0 / m_rr;
		}

		m_etq = (5.67e-6 * m_ens - 2.32e-3) * m_ens + 0.031;
		m_h0s = -15.0;

		ascat = 0.0;
	} else {
		// Computer scatter attenuation - 24
		double h0;

		if (m_h0s > 15.0) {
			h0 = m_h0s;
		} else {
			double th = m_the[0] + m_the[1] + d * m_gme;
			double r2 = 2.0 * m_wn * th;
			double r1 = r2 * m_he[0];
			r2 *= m_he[1];

			if (r1 < 0.2 && r2 < 0.2) {
				ascat = 1001.0;
				goto noscat;
			}

			double ss = (d - m_ad) / (d + m_ad);
			double q  = m_rr / ss;
			ss = ::max(0.1, ss);
			q  = ::min(::max(0.1, q), 10.0);

			double z0  = (d - m_ad) * (d + m_ad) * th * 0.25 / d;
			double et  = (m_etq * ::exp(-::pow(::min(1.7, z0 / 8.0e3), 6.0)) + 1.0) * z0 / 1.7556e3;
			double ett = ::max(et, 1.0);

			h0  = (h0f(r1, ett) + h0f(r2, ett)) * 0.5;
			h0 += ::min(h0, (1.38 - ::log(ett)) * ::log(ss) * ::log(q) * 0.49);
			h0  = FORTRAN_DIM(h0, 0.0);

			if (et < 1.0) {
				double temp = (1.0 + 1.4142 / r1) * (1.0 + 1.4142 / r2);
				h0 = et * h0 + (1.0 - et) * 4.343 * ::log((temp * temp) * (r1 + r2) / (r1 + r2 + 2.8284));
			}

			if (h0 > 15.0 && m_h0s >= 0.0)
				h0 = m_h0s;
		}

		m_h0s = h0;
		double th = m_tha + d * m_gme;

		ascat = ahd(th * d) + 4.343 * ::log(47.7 * m_wn * th * th * th * th) - 0.1 * (m_ens - 301.0) * ::exp(-th * d / 40e3) + h0;
	}

noscat:
	return ascat;
}

/*
 * The inverse standard normal complementary probability function.
 *
 * Checked again ITM.pdf, 25.2.2005
 */
double CITM::qerfi(double q) const
{
	const double c0 = 2.515516698;
	const double c1 = 0.802853;
	const double c2 = 0.010328;
	const double d1 = 1.432788;
	const double d2 = 0.189269;
	const double d3 = 0.001308;

	double x = 0.5 - q;
	double t = ::max(0.5 - ::fabs(x), 0.000001);
	t = ::sqrt(-2.0 * ::log(t));

	double res = t - ((c2 * t + c1) * t + c0) / (((d3 * t + d2) * t + d1) * t + 1.0);

	return (x < 0.0) ? -res : res;
}

/*
 * Convert parameters ready for lrm_
 *
 * Checked against ITM.pdf, 25.2.2004
 */
void CITM::qlrps(double zsys, double en0, int ipol, double eps, double sgm)
{
	const double gma = 157e-9;

	m_wn  = double(m_f) / 47.7;
	m_ens = en0;

	if (zsys != 0.0)
		m_ens *= ::exp(-zsys / 9460.0);

	m_gme = gma * (1.0 - 0.04665 * ::exp(m_ens / 179.3));
	complex<double> zq(eps, 376.62 * sgm / m_wn);

	m_zgnd = sqrt(zq - 1.0);

	if (ipol != 0)
		m_zgnd /= zq;
}

/*
 * Function, abq.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::abq(const complex<double>& r) const
{
	return r.real() * r.real() + r.imag() * r.imag();
}

/*
 * Calculate the line-of-sight attenuation.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::losAttenuation(double d)
{
	double alos = 0.0;

	if (d == 0.0) {
		// Prepare initial los constants - 18
		m_wls = 0.021 / (0.021 + m_wn * m_dh / ::max(10e3, m_dlsa));
	} else {
		// Compute los attenuation - 19
		double q = (1.0 - 0.8 * ::exp(-d / 50e3)) * m_dh;
		double s = 0.78 * q * ::exp(-::pow(q / 16.0, 0.25));

		q = m_he[0] + m_he[1];

		double sps = q / ::sqrt(d * d + q * q);

		complex<double> r = (sps - m_zgnd) / (sps + m_zgnd) * ::exp(-::min(10.0, m_wn * s * sps));
		q = abq(r);

		if (q < 0.25 || q < sps)
			r *= ::sqrt(sps / q);

		alos = m_emd * d + m_aed;

		q = m_wn * m_he[0] * m_he[1] * 2.0 / d;

		if (q > M_PI_2)
			q = M_PI - 2.4649 / q;

		alos = (-4.343 * ::log(abq(complex<double>(::cos(q), -::sin(q)) + r)) - alos) * m_wls + alos;
	}

	return alos;
}

/*
 * The Longley-Rice propagation model.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
void CITM::lrProp(double d)
{
	// Do secondary parameters - 6
	for (int j = 0; j < 2; j++)
		m_dls[j] = ::sqrt(2.0 * m_he[j] / m_gme);

	m_dlsa = m_dls[0] + m_dls[1];
	m_dla  = m_dl[0] + m_dl[1];
	m_tha  = ::max(m_the[0] + m_the[1], -m_dla * m_gme);
	bool wlos  = false;
	bool wscat = false;

	// Check parameter ranges - 7
	if (m_wn < 0.838 || m_wn > 210.0)
		m_kwx = ::max(m_kwx, 1);

	for (int j = 0; j < 2; j++)
		if (m_hg[j] < 1.0 || m_hg[j] > 1000.0)
			m_kwx = ::max(m_kwx, 1);

	for (int j = 0; j < 2; j++)
		if (::fabs(m_the[j]) > 200e-3 || m_dl[j] < 0.1 * m_dls[j] || m_dl[j] > 3.0 * m_dls[j] )
			m_kwx = ::max(m_kwx, 3);

	if (m_ens < 250.0 || m_ens > 400.0 || m_gme < 75e-9 || m_gme > 250e-9 || m_zgnd.real() <= ::fabs(m_zgnd.imag()) || m_wn < 0.419 || m_wn > 420.0)
		m_kwx = 4;

	for (int j = 0; j < 2; j++)
		if (m_hg[j] < 0.5 || m_hg[j] > 3000.0)
			m_kwx = 4;

	double dmin = ::fabs(m_he[0] - m_he[1]) / 200e-3;

	// Diffraction coefficients - 9
	diffractionAttenuation(0.0);

	double xae  = ::pow(m_wn * m_gme * m_gme, -THIRD);

	double d3 = ::max(m_dlsa, 1.3787 * xae + m_dla);
	double d4 = d3 + 2.7574 * xae;

	double a3 = diffractionAttenuation(d3);
	double a4 = diffractionAttenuation(d4);

	m_emd = (a4 - a3)/ (d4 - d3);
	m_aed = a3 - m_emd * d3;

	// Check distance - 8
	if (m_dist > 0.0) {
		if (m_dist > 1000e3)
			m_kwx = ::max(m_kwx, 1);

		if (m_dist < dmin)
			m_kwx = ::max(m_kwx, 3);

		if (/* m_dist < 1e3 || */ m_dist > 2000e3)
			m_kwx = 4;
	}

	// LOS calculations - 15
	if (m_dist < m_dlsa) {
		if (!wlos) {
			// LOS coefficient - 16
			double  q = losAttenuation(0.0);
			double d2 = m_dlsa;
			double a2 = m_aed + d2 * m_emd;
			double d0 = 1.908 * m_wn * m_he[0] * m_he[1];

			double d1;
			if (m_aed >= 0.0) {
				d0 = ::min(d0, 0.5 * m_dla);
				d1 = d0 + 0.25 * (m_dla - d0);
			} else {
				d1 = ::max(-m_aed / m_emd, 0.25 * m_dla);
			}

			double a1 = losAttenuation(d1);
			bool wq = false;

			if (d0 < d1) {
			 	double a0 = losAttenuation(d0);
				double q  = ::log(d2 / d0);
				m_ak2 = ::max(0.0, ((d2 - d0) * (a1 - a0) - (d1 - d0) * (a2 - a0)) / ((d2 - d0) * ::log(d1 / d0) - (d1 - d0) * q));

				wq = m_aed > 0.0 || m_ak2 > 0.0;

				if (wq) { 
					m_ak1 = (a2 - a0 - m_ak2 * q) / (d2 - d0);

					if (m_ak1 < 0.0) {
						m_ak1 = 0.0;
						m_ak2 = FORTRAN_DIM(a2, a0) / q;

						if (m_ak2 == 0.0)
							m_ak1 = m_emd;
					}
				}
			}

			if (!wq) {
				m_ak1 = FORTRAN_DIM(a2, a1) / (d2 - d1);
				m_ak2 = 0.0;

				if (m_ak1 == 0.0)
					m_ak1 = m_emd;
			}

			m_ael = a2 - m_ak1 * d2 - m_ak2 * ::log(d2);

			wlos = true;
		}

		if (m_dist > 0.0)
			m_aref = m_ael + m_ak1 * m_dist + m_ak2 * ::log(m_dist);
	}

	// Troposcatter calculations - 20
	if (m_dist <= 0.0 || m_dist >= m_dlsa) {
		if (!wscat) {
			// Troposcatter coefficients - 21
			double q  = scatterAttenuation(0.0);
			double d5 = m_dla + 200e3;
			double d6 = d5 + 200e3;
			double a6 = scatterAttenuation(d6);
			double a5 = scatterAttenuation(d5);

			if (a5 < 1000.0) {
				m_ems = (a6 - a5) / 200e3;
				m_dx  = ::max(m_dlsa, ::max(m_dla + 0.3 * xae * ::log(47.7 * m_wn), (a5 - m_aed - m_ems * d5) / (m_emd - m_ems)));
				m_aes = (m_emd - m_ems) * m_dx + m_aed;
			} else {
				m_ems = m_emd;
				m_aes = m_aed;
				m_dx  = 10e6;
			}

			wscat = true;
		}

		if (m_dist > m_dx)
			m_aref = m_aes + m_ems * m_dist;
		else
			m_aref = m_aed + m_emd * m_dist;
	}

	m_aref= ::max(m_aref, 0.0);
}

/*
 * Function, curve.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::curve(double c1, double c2, double x1, double x2, double x3, double de) const
{
	double temp1= (de - x2) / x3;
	double temp2 = de / x1;

	temp1 *= temp1;
	temp2 *= temp2;

	return (c1 + c2 / (1.0 + temp1)) * temp2 / (1.0 + temp2);
}

/*
 * Calculate two/threefold quantile of attenuation.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::avar()
{
	const double bv1[7]  = {-9.67, -0.62, 1.26, -9.21, -0.62, -0.39, 3.15};
	const double bv2[7]  = {12.7, 9.19, 15.5, 9.05, 9.19, 2.86, 857.9};
	const double xv1[7]  = {144.9e3, 228.9e3, 262.6e3, 84.1e3, 228.9e3, 141.7e3, 2222.e3};
	const double xv2[7]  = {190.3e3, 205.2e3, 185.2e3, 101.1e3, 205.2e3, 315.9e3, 164.8e3};
	const double xv3[7]  = {133.8e3, 143.6e3, 99.8e3, 98.6e3, 143.6e3, 167.4e3, 116.3e3};
	const double bsm1[7] = {2.13, 2.66, 6.11, 1.98, 2.68, 6.86, 8.51};
	const double bsm2[7] = {159.5, 7.67, 6.65, 13.11, 7.16, 10.38, 169.8};
	const double xsm1[7] = {762.2e3, 100.4e3, 138.2e3, 139.1e3, 93.7e3, 187.8e3, 609.8e3};
	const double xsm2[7] = {123.6e3, 172.5e3, 242.2e3, 132.7e3, 186.8e3, 169.6e3, 119.9e3};
	const double xsm3[7] = {94.5e3, 136.4e3, 178.6e3, 193.5e3, 133.5e3, 108.9e3, 106.6e3};
	const double bsp1[7] = {2.11, 6.87, 10.08, 3.68, 4.75, 8.58, 8.43};
	const double bsp2[7] = {102.3, 15.53, 9.60, 159.3, 8.12, 13.97, 8.19};
	const double xsp1[7] = {636.9e3, 138.7e3, 165.3e3, 464.4e3, 93.2e3, 216.0e3, 136.2e3};
	const double xsp2[7] = {134.8e3, 143.7e3, 225.7e3, 93.1e3, 135.9e3, 152.0e3, 188.5e3};
	const double xsp3[7] = {95.6e3, 98.6e3, 129.7e3, 94.2e3, 113.4e3, 122.7e3, 122.9e3};
	const double bsd1[7] = {1.224, 0.801, 1.380, 1.000, 1.224, 1.518, 1.518};
	const double bzd1[7] = {1.282, 2.161, 1.282, 20., 1.282, 1.282, 1.282};
	const double bfm1[7] = {1.0, 1.0, 1.0, 1.0, 0.92, 1.0, 1.0};
	const double bfm2[7] = {0.0, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0};
	const double bfm3[7] = {0.0, 0.0, 0.0, 0.0, 1.77, 0.0, 0.0};
	const double bfp1[7] = {1.0, 0.93, 1.0, 0.93, 0.93, 1.0, 1.0};
	const double bfp2[7] = {0.0, 0.31, 0.0, 0.19, 0.31, 0.0, 0.0};
	const double bfp3[7] = {0.0, 2.00, 0.0, 1.79, 2.00, 0.0, 0.0};
	const double rt = 7.8;

	// Climactic coefficients - 32
	double cv1  = bv1[m_klim];
	double cv2  = bv2[m_klim];
	double yv1  = xv1[m_klim];
	double yv2  = xv2[m_klim];
	double yv3  = xv3[m_klim];
	double csm1 = bsm1[m_klim];
	double csm2 = bsm2[m_klim];
	double ysm1 = xsm1[m_klim];
	double ysm2 = xsm2[m_klim];
	double ysm3 = xsm3[m_klim];
	double csp1 = bsp1[m_klim];
	double csp2 = bsp2[m_klim];
	double ysp1 = xsp1[m_klim];
	double ysp2 = xsp2[m_klim];
	double ysp3 = xsp3[m_klim];
	double csd1 = bsd1[m_klim];
	double zd   = bzd1[m_klim];
	double cfm1 = bfm1[m_klim];
	double cfm2 = bfm2[m_klim];
	double cfm3 = bfm3[m_klim];
	double cfp1 = bfp1[m_klim];
	double cfp2 = bfp2[m_klim];
	double cfp3 = bfp3[m_klim];

	// Frequency coefficients - 34
	double q  = ::log(0.133 * m_wn);

	double gm = cfm1 + cfm2 / ((cfm3 * q * cfm3 * q) + 1.0);
	double gp = cfp1 + cfp2 / ((cfp3 * q * cfp3 * q) + 1.0);

	// System coefficients - 35
	double dexa = ::sqrt(18e6 * m_he[0]) + ::sqrt(18e6 * m_he[1]) + ::pow((575.7e12 / m_wn), THIRD);

	// Distance coefficients - 36
	double de;
	if (m_dist < dexa)
		de = 130e3 * m_dist / dexa;
	else
		de = 130e3 + m_dist - dexa;

	double vmd  = curve(cv1, cv2, yv1, yv2, yv3, de);
	double sgtm = curve(csm1, csm2, ysm1, ysm2, ysm3, de) * gm;
	double sgtp = curve(csp1, csp2, ysp1, ysp2, ysp3, de) * gp;
	double sgtd = sgtp * csd1;
	double tgtd = (sgtp - sgtd) * zd;

	double vs0 = ::pow(5.0 + 3.0 * ::exp(-de / 100e3), 2.0);

	// Correct normal deviates - 37
	if (::fabs(m_zc) > 3.1)
		m_kwx = ::max(m_kwx, 1);

	// Resolve standard deviations - 38
	double sgt;
	if (m_zc < 0.0)
		sgt = sgtm;
	else if (m_zc <= zd)
		sgt = sgtp;
	else
		sgt = sgtd + tgtd / m_zc;

	double vs = vs0 + (sgt * m_zc * sgt * m_zc) / (rt + m_zc * m_zc);

	// Resolve deviations, yr, yc - 39
	m_sgc = ::sqrt(sgt * sgt + vs);

	double avar = m_aref - vmd - m_sgc * m_zc;

	if (avar < 0.0)
		avar *= (29.0 - avar) / (29.0 - 10.0 * avar);

	return avar;
}

/*
 * Find the horizons for the two end points.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
void CITM::horizons(double pfl[])
{
	wxASSERT(pfl != NULL);

	int    np = int(pfl[0]);
	double xi = pfl[1];
	double za = pfl[2] + m_hg[0];
	double zb = pfl[np + 2] + m_hg[1];
	double qc = 0.5 * m_gme;
	double q  = qc * m_dist;

	m_the[1] = (zb - za) / m_dist;
	m_the[0] = m_the[1] - q;
	m_the[1] = -m_the[1] - q;
	m_dl[0]  = m_dist;
	m_dl[1]  = m_dist;

	if (np >= 2) {
		double sa = 0.0;
		double sb = m_dist;
		bool   wq = true;

		for (int i = 1; i < np; i++) {
			sa += xi;
			sb -= xi;
			q   = pfl[i + 2] - (qc * sa + m_the[0]) * sa - za;

			if (q > 0.0) {
				m_the[0] += q / sa;
				m_dl[0]   = sa;
				wq = false;
			}

			if (!wq) {
				q = pfl[i + 2] - (qc * sb + m_the[1]) * sb - zb;

				if (q > 0.0) {
					m_the[1] += q / sb;
					m_dl[1]   = sb;
				}
			}
		}
	}
}
  
/*
 * Least squares fit.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
void CITM::lsqf(double z[], double x1, double x2, double& z0, double& zn) const
{
	wxASSERT(z != NULL);

	double xn = z[0];
	double xa = int(FORTRAN_DIM(x1 / z[1], 0.0));
	double xb = xn - int(FORTRAN_DIM(xn, x2 / z[1]));

	if (xb <= xa) {
		xa = FORTRAN_DIM(xa, 1.0);
		xb = xn - FORTRAN_DIM(xn, xb + 1.0);
	}

	int ja = int(xa);
	int jb = int(xb);
	int  n = jb - ja;

	xa = xb - xa;
	double x = -0.5 * xa;

	xb += x;
	double a = 0.5 * (z[ja + 2] + z[jb + 2]);
	double b = 0.5 * (z[ja + 2] - z[jb + 2]) * x;

	for (int i = 2; i <= n; i++) {
		ja++;
		x += 1.0;
		a += z[ja + 2];
		b += z[ja + 2] * x;
	}

	a /= xa;
	b *= 12.0 / ((xa * xa + 2.0) * xa);

	z0 = a - b * xb;
	zn = a + b * (xn - xb);
}

/*
 * Calculates the quantile.
 *
 * Checked against ITM.pdf, 25.3.2004
 * Not very happy with it, but it works.
 */
double CITM::quantile(int nn, double a[], int ir) const
{
	wxASSERT(a != NULL);

	int m = 0;
	int n = nn;
	int k = ::min(::max(0, ir), n);

	bool done   = false;
	bool goto10 = true;

	double q = 0.0;
	int j1 = 0;
	int i0 = 0;

	while (!done) {
		if (goto10) {
			q = a[k];
			i0 = m;
			j1 = n;
		}

		int i = i0;

		while (i <= n && a[i] >= q)
			i++;

		if (i > n)
			i = n;

		int j = j1;

		while (j >= m && a[j] <= q)
			j--;

		if (j < m)
			j = m;

		if (i < j) {
			double r = a[i];
			a[i] = a[j];
			a[j] = r;

			i0 = i + 1;
			j1 = j - 1;
			goto10 = false;
		} else if (i < k) {
			a[k] = a[i];
			a[i] = q;
			m = i + 1;
			goto10 = true;
		} else if (j > k) {
			a[k] = a[j];
			a[j] = q;
			n = j - 1;
			goto10 = true;
		} else {
			done = true;
		}
	}

	return q;
}

/*
 * Find interdecile range of elevations.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
double CITM::dlthx(double pfl[], double x1, double x2) const
{
	wxASSERT(pfl != NULL);

	int np = int(pfl[0]);
	double xa = x1 / pfl[1];
	double xb = x2 / pfl[1];

	double dlthx = 0.0;

	if (xb - xa >= 2.0) {
		int ka = int(0.1 * (xb - xa + 8.0));
		ka = ::min(::max(4, ka), 25);
		int n  = 10 * ka - 5;
		int kb = n - ka + 1;
		double sn = double(n - 1);

		double* s = new double[n + 2];

		s[0] = sn;
		s[1] = 1.0;
		xb = (xb - xa) / sn;
		int k = int(xa + 1.0);
		xa -= double(k);

		for (int j = 0; j < n; j++) {
			while (xa > 0.0 && k < np) {
				xa -= 1.0;
				k++;
			}

			s[j + 2] = pfl[k + 2] + (pfl[k + 2] - pfl[k + 1]) * xa;
			xa += xb;
		}

		lsqf(s, 0.0, sn, xa, xb);
		xb = (xb - xa) / sn;

		for (int j = 0; j < n; j++) {
			s[j + 2] -= xa;
			xa += xb;
		}

		dlthx = quantile(n - 1, s + 2, ka - 1) - quantile(n - 1, s + 2, kb - 1);
		dlthx /= 1.0 - 0.8 * ::exp(-(x2 - x1) / 50.0e3);

		delete[] s;
	}

	return dlthx;
}

/*
 * Prepare for point-to-point mode.
 *
 * Checked against ITM.pdf, 25.2.2004
 */
void CITM::qlrpfl(double pfl[])
{
	wxASSERT(pfl != NULL);

	m_dist = pfl[0] * pfl[1];
	int np = int(pfl[0]);

	double xl[2];

	// Horizons and dh from pfl - 44
	horizons(pfl);

	for (int j = 0; j < 2; j++)
		xl[j] = ::min(15.0 * m_hg[j], 0.1 * m_dl[j]);

	xl[1] = m_dist - xl[1];
	m_dh = dlthx(pfl, xl[0], xl[1]);

	if (m_dl[0] + m_dl[1] > 1.5 * m_dist) {
		// Re-do LOS horizons - 45
		double za, zb;
		lsqf(pfl, xl[0], xl[1], za, zb);

		m_he[0] = m_hg[0] + FORTRAN_DIM(pfl[2], za);
		m_he[1] = m_hg[1] + FORTRAN_DIM(pfl[np + 2], zb);

		for (int j = 0; j < 2; j++)
			m_dl[j] = ::sqrt(2.0 * m_he[j] / m_gme) * ::exp(-0.07 * ::sqrt(m_dh / ::max(m_he[j], 5.0)));

		double q = m_dl[0] + m_dl[1];

		if (q <= m_dist) {
			q = ::pow(m_dist / q, 2.0);

			for (int j = 0; j < 2; j++) {
				m_he[j] *= q;
				m_dl[j]  = ::sqrt(2.0 * m_he[j] / m_gme) * ::exp(-0.07 * ::sqrt(m_dh / ::max(m_he[j], 5.0)));
			}
		}

		for (int j = 0; j < 2; j++) {
			double q = ::sqrt(2.0 * m_he[j] / m_gme);
			m_the[j] = (0.65 * m_dh * (q / m_dl[j] - 1.0) - 2.0 * m_he[j]) / q;
		}
	} else {
		// Transhorizon effective heights - 46
		double za, zb, q;
		lsqf(pfl, xl[0], 0.9 * m_dl[0], za, q);
		lsqf(pfl, m_dist - 0.9 * m_dl[1], xl[1], q, zb);

		m_he[0] = m_hg[0] + FORTRAN_DIM(pfl[2], za);
		m_he[1] = m_hg[1] + FORTRAN_DIM(pfl[np + 2], zb);
	}

	lrProp(0.0);
}

/******************************************************************************

	pol:
		0-Horizontal, 1-Vertical

	radio_climate:
		0-Equatorial, 1-Continental Subtropical,
		2-Maritime Tropical, 3-Desert, 4-Continental Temperate,
		5-Maritime Temperate, Over Land, 6-Maritime Temperate,
		Over Sea

	conf, rel: .01 to .99

	elev[]: [num points - 1], [delta dist(meters)],
	        [height(meters) point 1], ..., [height(meters) point n]

	errnum: 0- No Error.
		1- Warning: Some parameters are nearly out of range.
		            Results should be used with caution.
		2- Note: Default parameters have been substituted for
		         impossible ones.
		3- Warning: A combination of parameters is out of range.
			    Results are probably invalid.
		Other-  Warning: Some parameters are out of range.
			Results are probably invalid.

*****************************************************************************/
void CITM::calculate(double elev[], int txHeight, int rxHeight, double &dbloss, wxString& strmode)
{
	wxASSERT(elev != NULL);
	wxASSERT(txHeight >= 0);
	wxASSERT(rxHeight >= 0);

	strmode.Clear();

	m_hg[0] = double(txHeight);
	m_hg[1] = double(rxHeight);
	m_kwx   = 0;

	qlrpfl(elev);

	double fs = 32.45 + 20.0 * ::log10(double(m_f)) + 20.0 * ::log10(m_dist / 1000.0);

	int q = int(m_dist - m_dla);

	if (q < 0) {
		strmode.Append(wxT("Line of Sight"));
	} else {
		if (q == 0)
			strmode.Append(wxT("Single Horizon"));
		else
			strmode.Append(wxT("Double Horizon"));

		if (m_dist <= m_dlsa || m_dist <= m_dx)
			strmode.Append(wxT(", Diffraction Dominant"));
		else if (m_dist > m_dx)
			strmode.Append(wxT(", Troposcatter Dominant"));
	}

	dbloss = avar() + fs;
}

int CITM::getErr() const
{
	return m_kwx;
}
