/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Path.h"
#include "Defs.h"

CPath::CPath(const CLocation& source) :
m_slat(0.0),
m_slon(0.0),
m_dlat(0.0),
m_dlon(0.0),
m_azimuth(0.0),
m_distance(0.0)
{
	m_slat = source.getLatitude();
	m_slon = source.getLongitude();
}

CPath::CPath(double slat, double slon) :
m_slat(0.0),
m_slon(0.0),
m_dlat(0.0),
m_dlon(0.0),
m_azimuth(0.0),
m_distance(0.0)
{
	m_slat = slat;
	m_slon = slon;
}

CPath::~CPath()
{
}

void CPath::setDest(const CLocation& location)
{
	m_dlat = location.getLatitude();
	m_dlon = location.getLongitude();

	fillinData();
}

void CPath::setDest(double lat, double lon)
{
	m_dlat = lat;
	m_dlon = lon;

	fillinData();
}

void CPath::setDestAzDist(double az, double dist)
{
	m_azimuth  = az;
	m_distance = dist;

	double azr   = RAD(az);
	double distr = dist / R;

	double colat = M_PI_2 - RAD(m_slat);
	double cosco = ::cos(colat);
	double sinco = ::sin(colat);
	double cosgc = ::cos(distr);
	double singc = ::sin(distr);

	double cosb  = cosco * cosgc + sinco * singc * ::cos(azr);
	double arg   = ::max(0.0, (1.0 - cosb * cosb));
	double b     = ::atan2(::sqrt(arg), cosb);
	double arc   = (cosgc - cosco * cosb) / (sinco * ::sin(b));

	arg   = ::max(0.0, (1.0 - arc * arc));
	double rdlon = ::atan2(::sqrt(arg), arc);

	double rlat = M_PI_2 - ::fabs(b);

	m_dlat = DEG(FORTRAN_SIGN(rlat, cosb));

	if (azr <= M_PI)
		m_dlon = m_slon + DEG(::fabs(rdlon));
	else
		m_dlon = m_slon - DEG(::fabs(rdlon));
}

void CPath::fillinData()
{
	double sinslat = ::sin(RAD(m_slat));
	double cosslat = ::cos(RAD(m_slat));
	double sindlat = ::sin(RAD(m_dlat));
	double cosdlat = ::cos(RAD(m_dlat));
	double coslon  = ::cos(RAD(m_slon - m_dlon));

	double cosdist = sinslat * sindlat + cosslat * cosdlat * coslon;

	if (cosdist > 1.0)
		cosdist = 1.0;
	else if (cosdist < -1.0)
		cosdist = -1.0;

	double sindist = ::sqrt(1.0 - cosdist * cosdist);

	double distance = ::acos(cosdist);

	double cosaz = (sindlat - sinslat * cosdist) / (cosslat * sindist);

	if (cosaz > 1.0)
		cosaz = 1.0;
	else if (cosaz < -1.0)
		cosaz = -1.0;

	double azimuth = ::acos(cosaz);

	double diff = RAD(m_dlon - m_slon);

	if (diff <= -M_PI)
		diff += (2.0 * M_PI);

	if (diff >= M_PI)
		diff -= (2.0 * M_PI);

	if (diff < 0.0)
		azimuth = 2.0 * M_PI - azimuth;

	m_distance = distance * R;
	m_azimuth  = DEG(azimuth);
}

CLocation CPath::getStart() const
{
	return CLocation(m_slat, m_slon);
}

CLocation CPath::getEnd() const
{
	return CLocation(m_dlat, m_dlon);
}

double CPath::getDistance() const
{
	return m_distance;
}

double CPath::getBearing() const
{
	return m_azimuth;
}

void CPath::getPath(double distance, double& latitude, double& longitude) const
{
	wxASSERT(distance >= 0.0);

	double latSlope = (m_dlat - m_slat) / m_distance;
	double lonSlope = (m_dlon - m_slon) / m_distance;

	latitude  = m_slat + latSlope * distance;
	longitude = m_slon + lonSlope * distance;
}

CLocation CPath::getPath(double distance) const
{
	wxASSERT(distance >= 0.0);

	double latSlope = (m_dlat - m_slat) / m_distance;
	double lonSlope = (m_dlon - m_slon) / m_distance;

	double latitude  = m_slat + latSlope * distance;
	double longitude = m_slon + lonSlope * distance;

	return CLocation(latitude, longitude);
}

double CPath::FORTRAN_SIGN(double a, double b) const
{
	double sgn = (b >= 0.0) ? 1.0 : -1.0;

	return ::fabs(a) * sgn;
}
