/*
 *   Copyright (C) 2003,2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "SRTM.h"
#include "Defs.h"

#include <wx/log.h>
#include <wx/filename.h>
#include <wx/file.h>
#include <wx/zipstrm.h>
#include <wx/wfstream.h>

const int N_ROWS = 1201;
const int N_COLS = 1201;

const double LAT_INCR = 1.0 / double(N_ROWS - 1);
const double LON_INCR = 1.0 / double(N_COLS - 1);

CSRTM::CSRTM(const wxString& path, double latitude, double longitude) :
CDEM(),
m_minLatitude(0),
m_maxLatitude(0),
m_minLongitude(0),
m_maxLongitude(0),
m_llX(longitude),
m_llY(latitude),
m_data(NULL),
m_checked(NULL)
{
	wxString baseName = findFileName(m_llY, m_llX);

	wxString file = baseName + wxT(".hgt");
	wxString name = baseName + wxT(".hgt");

	bool ret = readZipFileData(path, name, wxT("zip"), file);

	if (!ret) {
		name = baseName + wxT("_hgt");

		ret = readZipFileData(path, name, wxT("zip"), file);

		if (!ret)
			readRawFileData(path, baseName, wxT("hgt"));
	}
}

CSRTM::~CSRTM()
{
	if (m_data != NULL)
		delete[] m_data;

	if (m_checked != NULL)
		delete[] m_checked;
}

int CSRTM::getHeight(double latitude, double longitude)
{
	wxASSERT(latitude >= m_minLatitude);
	wxASSERT(latitude < m_maxLatitude);
	wxASSERT(longitude >= m_minLongitude);
	wxASSERT(longitude < m_maxLongitude);

	if (m_data == NULL)
		return VOID_DATA;

	int row = N_ROWS - 1 - int(::fabs(m_llY - latitude)  / LAT_INCR);
	int col = int(::fabs(m_llX - longitude) / LON_INCR);

	wxASSERT(row >= 0 && row < N_ROWS);
	wxASSERT(col >= 0 && col < N_COLS);

	return readHeight(row, col);
}

void CSRTM::setHeight(int height, double latitude, double longitude)
{
	wxASSERT(latitude >= m_minLatitude);
	wxASSERT(latitude < m_maxLatitude);
	wxASSERT(longitude >= m_minLongitude);
	wxASSERT(longitude < m_maxLongitude);

	if (m_data == NULL)
		return;

	int row = N_ROWS - 1 - int(::fabs(m_llY - latitude)  / LAT_INCR);
	int col = int(::fabs(m_llX - longitude) / LON_INCR);

	wxASSERT(row >= 0 && row < N_ROWS);
	wxASSERT(col >= 0 && col < N_COLS);

	int offset = row * N_COLS + col;

	m_data[offset]    = height;
	m_checked[offset] = true;
}

bool CSRTM::isInData(double latitude, double longitude) const
{
	return (latitude  >= double(m_minLatitude)  &&
		latitude  <  double(m_maxLatitude)  &&
		longitude >= double(m_minLongitude) &&
		longitude <  double(m_maxLongitude));
}

bool CSRTM::dataExists() const
{
	return m_data != NULL;
}

double CSRTM::getMinLatitude() const
{
	return double(m_minLatitude);
}

double CSRTM::getMaxLatitude() const
{
	return double(m_maxLatitude);
}

double CSRTM::getMinLongitude() const
{
	return double(m_minLongitude);
}

double CSRTM::getMaxLongitude() const
{
	return double(m_maxLongitude);
}

wxString CSRTM::findFileName(double& latitude, double& longitude)
{
	wxChar nsText;
	int    nsValue;

	if (latitude >= 0.0) {
		nsText = wxT('N');
		nsValue = int(latitude);

		m_minLatitude = nsValue;
		m_maxLatitude = m_minLatitude + 1;
	} else {
		nsText  = wxT('S');
		nsValue = int(latitude - 1.0);

		m_minLatitude = nsValue;
		m_maxLatitude = m_minLatitude + 1;
	}		

	latitude = double(nsValue);

	wxChar ewText;
	int    ewValue;

	if (longitude >= 0.0) {
		ewText = wxT('E');
		ewValue = int(longitude);

		m_minLongitude = ewValue;
		m_maxLongitude = m_minLongitude + 1;
	} else {
		ewText  = wxT('W');
		ewValue = int(longitude - 1.0);

		m_minLongitude = ewValue;
		m_maxLongitude = m_minLongitude + 1;
	}		

	longitude = double(ewValue);

	wxString fileName;
	fileName.Printf(wxT("%c%02d%c%03d"), nsText, ::abs(nsValue), ewText, ::abs(ewValue));

	return fileName;
}

bool CSRTM::readFileData(wxInputStream& stream)
{
	bool ret = stream.IsOk();

	if (!ret)
		return false;

	m_data    = new short[N_ROWS * N_COLS];
	m_checked = new bool[N_ROWS * N_COLS];

	for (int i = 0; i < (N_ROWS * N_COLS); i++) {
		wxInt16 data;
		stream.Read(&data, sizeof(wxInt16));

		m_data[i] = wxINT16_SWAP_ON_LE(data);
	}

	return true;
}

bool CSRTM::readZipFileData(const wxString& path, const wxString& name, const wxString& ext, const wxString& file)
{
	wxFileName fileName(path, name, ext);

	if (!fileName.IsOk())
		return false;

	if (!fileName.FileExists())
		return false;

	wxZipInputStream stream(fileName.GetFullPath(), file);

	return readFileData(stream);
}

bool CSRTM::readRawFileData(const wxString& path, const wxString& name, const wxString& ext)
{
	wxFileName fileName(path, name, ext);

	if (!fileName.IsOk())
		return false;

	if (!fileName.FileExists())
		return false;

	wxFFileInputStream stream(fileName.GetFullPath());

	return readFileData(stream);
}

int CSRTM::readHeight(int row, int col) const
{
	if (row < 0 || row >= N_ROWS ||
	    col < 0 || col >= N_COLS)
		return VOID_DATA;

	int offset = row * N_COLS + col;

	int height = m_data[offset];

	if (height == -32768)
		return VOID_DATA;

	if (height < -15)
		return 0;

	return height;
}
