/*
 *   Copyright (C) 2003,2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "SiteList.h"
#include "Defs.h"

#include <wx/listimpl.cpp>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/tokenzr.h>

WX_DEFINE_LIST(CSiteData);

CSiteList* CSiteList::m_instance = NULL;

CSiteList::CSiteList(const wxString& path) :
m_path(path),
m_siteData(NULL),
m_lat1(0.0),
m_lon1(0.0),
m_lat2(0.0),
m_lon2(0.0),
m_node(NULL)
{
	m_siteData = new CSiteData(wxKEY_NONE);
	m_siteData->DeleteContents(true);

	loadSites();
}

CSiteList::~CSiteList()
{
	m_siteData->DeleteContents(true);

	delete m_siteData;
}

CSiteList* CSiteList::create(const wxString& path)
{
	if (m_instance == NULL)
		m_instance = new CSiteList(path);

	return m_instance;
}

CSiteList* CSiteList::getInstance()
{
	return m_instance;
}

CPoint* CSiteList::find(const wxString& name) const
{
	CSiteData::Node* node = m_siteData->GetFirst();

	while (node != NULL) {
		CPoint* point = node->GetData();

		if (name.IsSameAs(point->getName(), false))
			return point;

		node = node->GetNext();
	}

	return NULL;
}

CPoint* CSiteList::getFirst(double lat1, double lon1, double lat2, double lon2)
{
	if (lat1 < lat2) {
		m_lat1 = lat1;
		m_lat2 = lat2;
	} else {
		m_lat1 = lat2;
		m_lat2 = lat1;
	}

	if (lon1 < lon2) {
		m_lon1 = lon1;
		m_lon2 = lon2;
	} else {
		m_lon1 = lon2;
		m_lon2 = lon1;
	}

	return getData(true);
}

CPoint* CSiteList::getNext()
{
	return getData(false);
}

CPoint* CSiteList::getData(bool first)
{
	if (first)
		m_node = m_siteData->GetFirst();
	else
		m_node = m_node->GetNext();

	while (m_node != NULL) {
		CPoint* site = m_node->GetData();

		double lat = site->getLocation().getLatitude();
		double lon = site->getLocation().getLongitude();

		if (lat >= m_lat1 && lat < m_lat2 &&
		    lon >= m_lon1 && lon < m_lon2)
			return site;

		m_node = m_node->GetNext();
	}

	return NULL;
}

int CSiteList::getCount() const
{
	return m_siteData->GetCount();
}

void CSiteList::loadSites()
{
	wxFileName fileName(m_path, wxT("sites.dat"));

	if (!fileName.IsOk())
		return;

	wxTextFile file;

	if (!file.Open(fileName.GetFullPath()))
		return;

	int n = file.GetLineCount();

	wxString name;
	double   latitude = double(VOID_DATA);

	for (int line = 0; line < n; line++) {
		wxString str = file.GetLine(line);

		if (str.GetChar(0) == wxT('[')) {
			int n = str.Find(wxT(']'));

			wxASSERT(n > 0);

			name = str.Mid(1, n - 1);
		} else if (str.Mid(0, 4).IsSameAs(wxT("lat="))) {
			str.Mid(4).ToDouble(&latitude);
		} else if (str.Mid(0, 4).IsSameAs(wxT("lon="))) {
			double longitude;
			str.Mid(4).ToDouble(&longitude);

			CLocation location(latitude, longitude);

			CPoint* site = new CPoint(name, location, 0, VOID_DATA);

			m_siteData->Insert(site);
		} else if (str.Mid(0, 4).IsSameAs(wxT("loc="))) {
			wxString locator = str.Mid(4);

			CLocation location(FORMAT_LOCATOR, locator);

			CPoint* site = new CPoint(name, location, 0, VOID_DATA);

			m_siteData->Insert(site);
		} else if (str.Mid(0, 4).IsSameAs(wxT("ngr="))) {
			wxString ngr = str.Mid(4);

			CLocation location(FORMAT_NGR, ngr);

			CPoint* site = new CPoint(name, location, 0, VOID_DATA);

			m_siteData->Insert(site);
		} else if (str.Mid(0, 4).IsSameAs(wxT("utm="))) {
			wxString utm = str.Mid(4);

			CLocation location(FORMAT_UTM, utm);

			CPoint* site = new CPoint(name, location, 0, VOID_DATA);

			m_siteData->Insert(site);
		}
	}

	file.Close();
}
