/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Terrain.h"
#include "Defs.h"
#include "ITM.h"
#include "App.h"

const int DATA_SIZE = 300;

const double K = 1.33;
const double SURFREF = 301.0;

BEGIN_EVENT_TABLE(CTerrain, wxPanel)
	EVT_MOTION(CTerrain::onMouseMove)
	EVT_PAINT(CTerrain::onPaint)
END_EVENT_TABLE()

CTerrain::CTerrain(wxWindow* parent, const CPoint& start, const CLocation& end, int endHeight, int endASL, int frequency, int polarisation, int climate, int ground, int confidence, int txPower, int txAntGain, int rxPower, int rxAntGain) :
wxPanel(parent, -1, wxDefaultPosition, wxSize(TERRAIN_WIDTH, TERRAIN_HEIGHT)),
m_bitmap(NULL),
m_path(start.getLocation()),
m_start(start),
m_end(end),
m_groundHeights(NULL),
m_spotHeights(NULL),
m_distances(NULL),
m_maxHeight(0),
m_count(0),
m_startHeight(0),
m_endHeight(0),
m_txHeight(0),
m_rxHeight(0),
m_frequency(frequency),
m_polarisation(polarisation),
m_txPower(txPower),
m_txAntGain(txAntGain),
m_rxPower(rxPower),
m_rxAntGain(rxAntGain),
m_climate(climate),
m_ground(ground),
m_confidence(confidence)
{
	m_bitmap = new wxBitmap(TERRAIN_WIDTH, TERRAIN_HEIGHT);

	m_path.setDest(end);

	CDEMCache* cache = ::wxGetApp().getDEMCache();
	wxASSERT(cache != NULL);

	m_distances     = new double[DATA_SIZE];
	m_groundHeights = new int[DATA_SIZE];
	m_spotHeights   = new int[DATA_SIZE];

	if (m_start.getASL() != VOID_DATA) {
		m_startHeight = m_start.getASL();
	} else {
		double startLat = m_start.getLocation().getLatitude();
		double startLon = m_start.getLocation().getLongitude();

		m_startHeight = cache->getHeight(startLat, startLon, true);

		if (m_startHeight == VOID_DATA)
			m_startHeight = 0;
	}

	m_txHeight = m_start.getHeight();

	if (endASL != VOID_DATA) {
		m_endHeight = endASL;
	} else {
		double endLat = m_end.getLatitude();
		double endLon = m_end.getLongitude();
		m_endHeight = cache->getHeight(endLat, endLon, true);

		if (m_endHeight == VOID_DATA)
			m_endHeight = 0;
	}

	m_rxHeight = endHeight;

	drawTerrain();
}

CTerrain::~CTerrain()
{
	delete m_bitmap;

	delete[] m_distances;
	delete[] m_groundHeights;
	delete[] m_spotHeights;
}

void CTerrain::drawTerrain()
{
	m_maxHeight = 0;
	m_count     = 0;

	CDEMCache* cache = ::wxGetApp().getDEMCache();

	wxASSERT(cache != NULL);

	double distance = m_path.getDistance();
	double bearing  = m_path.getBearing();

	double step = distance / double(DATA_SIZE);

	double re = R * K;

	int y0 = int(1000.0 * ::sqrt(::pow(re, 2.0) - ::pow(distance / 2.0, 2.0)));

	double maxTheta1 = -9999.0; int max1 = 0;
	double maxTheta2 = -9999.0; int max2 = 0;

	double* elev = new double[DATA_SIZE + 2];

	int zsys = 0;

	for (m_count = 0; m_count < DATA_SIZE; m_count++) {
		m_distances[m_count] = double(m_count) * step;

		double startDistance = m_distances[m_count];
		double endDistance   = distance - startDistance;

		double lat, lon;
		m_path.getPath(m_distances[m_count], lat, lon);

		int height = cache->getHeight(lat, lon, true);

		if (height == VOID_DATA)
			height = 0;

		zsys += height;

		if (m_count == 0) {
			m_spotHeights[m_count]   = m_startHeight;
			m_groundHeights[m_count] = 0;
			elev[m_count + 2]        = double(m_startHeight);
		} else if (m_count == (DATA_SIZE - 1)) {
			m_spotHeights[m_count]   = m_endHeight;
			m_groundHeights[m_count] = 0;
			elev[m_count + 2]        = double(m_endHeight);
		} else {
			m_spotHeights[m_count]   = height;
			m_groundHeights[m_count] = int(1000.0 * ::sqrt(::pow(re, 2.0) - ::pow(startDistance - distance / 2.0, 2.0))) - y0;
			elev[m_count + 2]        = double(height);
		}

		m_maxHeight = ::max(m_maxHeight, m_spotHeights[m_count]);

		if (m_count > 0) {
			double theta = DEG((m_spotHeights[m_count] - (m_startHeight + m_txHeight)) / 1000.0 / startDistance - startDistance / 2.0 / re);

			if (theta > maxTheta1) {
				maxTheta1 = theta;
				max1      = m_count;
			}
		}

		if (m_count < (DATA_SIZE - 1)) {
			double theta = DEG((m_spotHeights[m_count] - (m_endHeight + m_rxHeight))   / 1000.0 / endDistance - endDistance / 2.0 / re);

			if (theta > maxTheta2) {
				maxTheta2 = theta;
				max2      = m_count;
			}
		}
	}

	m_maxHeight += int(125.0 * ::pow(distance, 2.0) / re);

	zsys /= DATA_SIZE;

	elev[0] = double(DATA_SIZE - 1);
	elev[1] = step * 1000.0;

	CITM itm(zsys, SURFREF, m_frequency, m_climate, m_ground, m_polarisation, m_confidence);

	double dbloss;
	wxString info;
	itm.calculate(elev, m_txHeight, m_rxHeight, dbloss, info);

	int err = itm.getErr();

	delete[] elev;

	// Flood the graph area with black to start with
	wxMemoryDC dc;
	dc.SelectObject(*m_bitmap);
	dc.BeginDrawing();

	dc.SetBackground(*wxWHITE_BRUSH);
	dc.Clear();

	int lastHeight = m_spotHeights[0] + m_groundHeights[0];

	wxString text;
	text.Printf(wxT("Distance %.2f km   Bearing %.1f"), distance, bearing);

	wxCoord textWidth, textHeight;
	dc.GetTextExtent(text, &textWidth, &textHeight);

	dc.SetTextForeground(*wxBLACK);
	dc.DrawText(text, TERRAIN_WIDTH / 2 - textWidth / 2, TERRAIN_HEIGHT - 50);

	if (err == 0 || err == 1 || err == 2 || err == 3) {
		text.Printf(wxT("Path loss at %d MHz is %.1f dB, %s"), m_frequency, dbloss, info.c_str());
		dc.GetTextExtent(text, &textWidth, &textHeight);
		dc.DrawText(text, TERRAIN_WIDTH / 2 - textWidth / 2, 15);

		int eirp = m_txPower + m_txAntGain;

		double signal = double(eirp) - dbloss + double(m_rxAntGain);

		double snr = signal - double(m_rxPower);

		wxChar sign = (snr >= 0.0) ? wxT('+') : wxT('-');

		text.Printf(wxT("EIRP is %d dBm, received signal is %.0f dBm, signal margin is %c%.0f dB"), eirp, signal, sign, ::fabs(snr));
		dc.GetTextExtent(text, &textWidth, &textHeight);
		dc.DrawText(text, TERRAIN_WIDTH / 2 - textWidth / 2, 30);
	}

	int y = TERRAIN_HEIGHT - 75;

	text = m_start.getLocation().getLocation(FORMAT_LOCATOR);
	dc.GetTextExtent(text, &textWidth, &textHeight);
	dc.DrawText(text, 1 * TERRAIN_WIDTH / 8 - textWidth / 2, y);
	y += 15;

	if (m_start.getLocation().isValid(FORMAT_NGR)) {
		dc.DrawText(m_start.getLocation().getLocation(FORMAT_NGR), 1 * TERRAIN_WIDTH / 8 - textWidth / 2, y);
		y += 15;
	}

	text.Printf(wxT("%dm ASL"), m_spotHeights[0] + m_groundHeights[0] + m_txHeight);
	dc.DrawText(text, 1 * TERRAIN_WIDTH / 8 - textWidth / 2, y);

	y = TERRAIN_HEIGHT - 75;

	text = m_end.getLocation(FORMAT_LOCATOR);
	dc.GetTextExtent(text, &textWidth, &textHeight);
	dc.DrawText(text, 7 * TERRAIN_WIDTH / 8 - textWidth / 2, y);
	y += 15;

	if (m_end.isValid(FORMAT_NGR)) {
		dc.DrawText(m_end.getLocation(FORMAT_NGR), 7 * TERRAIN_WIDTH / 8 - textWidth / 2, y);
		y += 15;
	}

	text.Printf(wxT("%dm ASL"), m_spotHeights[m_count - 1] + m_groundHeights[m_count - 1] + m_rxHeight);
	dc.DrawText(text, 7 * TERRAIN_WIDTH / 8 - textWidth / 2, y);

	text.Printf(wxT("K (%.2f)"), K);
	dc.DrawText(text, TERRAIN_WIDTH - 70, TERRAIN_HEIGHT - 200);

	for (int i = 1; i < m_count; i++) {
		dc.SetPen(*wxBLACK_PEN);
		drawLine(dc, i - 1, 0, i, m_groundHeights[i]);

		drawLine(dc, i - 1, lastHeight, i, m_spotHeights[i] + m_groundHeights[i]);

		lastHeight = m_spotHeights[i] + m_groundHeights[i];
	}

	dc.SetPen(*wxBLACK_PEN);
	drawLine(dc, 0, 0, m_count - 1, 0);
	drawLine(dc, m_count - 1, 0, m_count - 1, m_endHeight);
	drawLine(dc, 0, 0, 0, m_startHeight);

	dc.SetPen(*wxGREEN_PEN);
	drawLine(dc, m_count - 1, m_endHeight, m_count - 1, m_endHeight + m_rxHeight);
	drawLine(dc, 0, m_startHeight, 0, m_startHeight + m_txHeight);

	if (max2 != 0) {
		wxPen pen(*wxRED, 1, wxSHORT_DASH);
		dc.SetPen(pen);
		drawLine(dc, 0, m_startHeight + m_txHeight, max1, m_groundHeights[max1] + m_spotHeights[max1]);
		drawLine(dc, m_count - 1, m_endHeight + m_rxHeight, max2, m_groundHeights[max2] + m_spotHeights[max2]);
	} else {
		double c = double((m_endHeight + m_rxHeight) - (m_startHeight + m_txHeight)) / double(m_count);

		int lastD = 0;
		int lastH = m_startHeight + m_txHeight;

		for (int i = 1; i < m_count; i++) {
			int d = getFresnel(1, m_distances[i], distance);
			int h = m_startHeight + m_txHeight + int(double(i) * c + 0.5);

			if ((i % 10) < 8) {
				dc.SetPen(*wxRED_PEN);
				drawLine(dc, i - 1, lastH, i, h);

				dc.SetPen(*wxGREY_PEN);
				drawLine(dc, i - 1, lastH + lastD, i, h + d);
				drawLine(dc, i - 1, lastH - lastD, i, h - d);
			}

			lastH = h;
			lastD = d;
		}
	}

	dc.EndDrawing();
}

void CTerrain::onPaint(const wxPaintEvent& event)
{
	wxPaintDC dc(this);

	dc.BeginDrawing();
	dc.DrawBitmap(*m_bitmap, 0, 0, false);
	dc.EndDrawing();
}

void CTerrain::onMouseMove(const wxMouseEvent& event)
{
	static int lastX = 0, lastY = 0;

	int x, y;
	event.GetPosition(&x, &y);

	if (::abs(x - lastX) > 2 || ::abs(y - lastY) > 2) {
		wxClientDC dc(this);
		dc.BeginDrawing();
		dc.DrawBitmap(*m_bitmap, 0, 0, false);
		printTheta(dc, x, y, true);
		dc.EndDrawing();

		lastX = x;
		lastY = y;
	}
}

void CTerrain::drawLine(wxDC& dc, int x1, int y1, int x2, int y2) const
{
	int nx1 = countToX(x1);
	int ny1 = heightToY(y1);

	int nx2 = countToX(x2);
	int ny2 = heightToY(y2);

	dc.DrawLine(nx1, ny1, nx2, ny2);
}

void CTerrain::printTheta(wxDC& dc, int mouseX, int mouseY, bool line)
{
	int  count = xToCount(mouseX);
	int height = yToHeight(mouseY);

	if (count == -1 || height == VOID_DATA)
		return;

	CLocation location = m_path.getPath(m_distances[count]);

	double startDistance = m_distances[count];
	double endDistance   = m_distances[DATA_SIZE - 1] - startDistance;

	dc.SetBrush(*wxWHITE_BRUSH);
	dc.SetPen(*wxWHITE_PEN);
	dc.SetTextForeground(*wxBLACK);

	wxString text;
	if (location.isValid(FORMAT_NGR))
		text.Printf(wxT("%.1f km  %s  %s  %dm ASL"), startDistance, location.getLocation(FORMAT_LOCATOR).c_str(), location.getLocation(FORMAT_NGR).c_str(), height - m_groundHeights[count]);
	else
		text.Printf(wxT("%.1f km  %s  %dm ASL"), startDistance, location.getLocation(FORMAT_LOCATOR).c_str(), height - m_groundHeights[count]);

	wxCoord textWidth, textHeight;
	dc.GetTextExtent(text, &textWidth, &textHeight);

	int x = TERRAIN_WIDTH;
	int y = TERRAIN_HEIGHT - 70;

	dc.DrawRectangle(200, y, 250, 5);
	dc.DrawText(text, x / 2 - textWidth / 2, y);

	double re = R * K;
	double theta1 = DEG((height - m_groundHeights[count] - (m_startHeight + m_txHeight)) / 1000.0 / startDistance - startDistance / 2.0 / re);
	double theta2 = DEG((height - m_groundHeights[count] - (m_endHeight + m_rxHeight))   / 1000.0 / endDistance   - endDistance / 2.0 / re);

	if (::fabs(theta1) > 9.99) theta1 = 9.99;
	if (::fabs(theta2) > 9.99) theta2 = 9.99;

	text.Printf(wxT(" %.2f "), theta1);
	x = 1 * TERRAIN_WIDTH / 16;
	y = TERRAIN_HEIGHT - 130;

	dc.DrawRectangle(x, y, 10, 5);
	dc.DrawText(text, x, y);

	text.Printf(wxT(" %.2f "), theta2);
	x = 14 * TERRAIN_WIDTH / 16;

	dc.DrawRectangle(x, y, 10, 5);
	dc.DrawText(text, x, y);

	if (line) {
		wxPen pen(*wxBLACK, 1, wxSHORT_DASH);
		dc.SetPen(pen);
		drawLine(dc, 0, m_startHeight + m_txHeight, count, height);
		drawLine(dc, m_count - 1, m_endHeight + m_rxHeight, count, height);
	}
}

int CTerrain::xToCount(int x) const
{
	int     span = m_count / 5 + m_count + m_count / 5;
	double scale = double(TERRAIN_WIDTH) / double(span);

	int count = int(double(x) / scale) - m_count / 5;

	return (count < 0 || count >= m_count) ? -1 : count;
}

int CTerrain::yToHeight(int y) const
{
	int     span = m_maxHeight / 2 + m_maxHeight * 2;
	double scale = double(TERRAIN_HEIGHT) / double(span);

	int height = int(double(TERRAIN_HEIGHT - y) / scale) - m_maxHeight / 2;

	return (height < -(m_maxHeight / 2) || height >= (m_maxHeight * 2)) ? VOID_DATA : height;
}

int CTerrain::countToX(int count) const
{
	if (count < 0 || count >= m_count)
		return -1;

	int     span = m_count / 5 + m_count + m_count / 5;
	double scale = double(TERRAIN_WIDTH) / double(span);

	return int(scale * double(m_count / 5 + count));
}

int CTerrain::heightToY(int height) const
{
	if (height < -(m_maxHeight / 2) || height >= (m_maxHeight * 2))
		return -1;

	int     span = m_maxHeight / 2 + m_maxHeight * 2;
	double scale = double(TERRAIN_HEIGHT) / double(span);

	return TERRAIN_HEIGHT - int(scale * double(m_maxHeight / 2 + height));
}

int CTerrain::getFresnel(int n, double d1, double distance) const
{
	wxASSERT(n > 0);
	wxASSERT(d1 <= distance);
	wxASSERT(distance > 0.0);

	double wl = 299.820 / double(m_frequency);
	double d2 = distance - d1;

	return int(::sqrt((double(n) * wl * d1 * 1000.0 * d2) / distance) + 0.5);
}
